import click, os
from bench.config.procfile import setup_procfile
from bench.config.supervisor import generate_supervisor_config
from bench.app import get_current_frappe_version, get_current_branch

def execute(bench_path):
	frappe_branch = get_current_branch('frappe', bench_path)
	frappe_version = get_current_frappe_version(bench_path)

	if not (frappe_branch=='develop' or frappe_version >= 7):
		# not version 7+
		# prevent running this patch
		return False

	click.confirm('\nThis update will remove Celery config and prepare the bench to use Python RQ.\n'
		'And it will overwrite Procfile and supervisor.conf.\n'
		'If you don\'t know what this means, type Y ;)\n\n'
		'Do you want to continue?',
		abort=True)

	setup_procfile(bench_path, yes=True)

	# if production setup
	if os.path.exists(os.path.join(bench_path, 'config', 'supervisor.conf')):
		generate_supervisor_config(bench_path, yes=True)



from bench.utils import get_program, exec_cmd, get_cmd_output, fix_prod_setup_perms, get_bench_name, find_executable, CommandFailedError
from bench.config.supervisor import generate_supervisor_config
from bench.config.nginx import make_nginx_conf
import os, subprocess

def setup_production(user, bench_path='.', yes=False):
	generate_supervisor_config(bench_path=bench_path, user=user, yes=yes)
	make_nginx_conf(bench_path=bench_path, yes=yes)
	fix_prod_setup_perms(bench_path, frappe_user=user)
	remove_default_nginx_configs()

	bench_name = get_bench_name(bench_path)
	nginx_conf = '/etc/nginx/conf.d/{bench_name}.conf'.format(bench_name=bench_name)

	supervisor_conf_extn = "ini" if is_centos7() else "conf"
	supervisor_conf = os.path.join(get_supervisor_confdir(), '{bench_name}.{extn}'.format(
		bench_name=bench_name, extn=supervisor_conf_extn))

	# Check if symlink exists, If not then create it.
	if not os.path.islink(supervisor_conf):
		os.symlink(os.path.abspath(os.path.join(bench_path, 'config', 'supervisor.conf')), supervisor_conf)

	if not os.path.islink(nginx_conf):
		os.symlink(os.path.abspath(os.path.join(bench_path, 'config', 'nginx.conf')), nginx_conf)

	reload_supervisor()

	if os.environ.get('NO_SERVICE_RESTART'):
		return

	reload_nginx()

def disable_production(bench_path='.'):
	bench_name = get_bench_name(bench_path)

	# supervisorctl
	supervisor_conf_extn = "ini" if is_centos7() else "conf"
	supervisor_conf = os.path.join(get_supervisor_confdir(), '{bench_name}.{extn}'.format(
		bench_name=bench_name, extn=supervisor_conf_extn))

	if os.path.islink(supervisor_conf):
		os.unlink(supervisor_conf)

	exec_cmd('sudo supervisorctl reread')
	exec_cmd('sudo supervisorctl update')

	# nginx
	nginx_conf = '/etc/nginx/conf.d/{bench_name}.conf'.format(bench_name=bench_name)

	if os.path.islink(nginx_conf):
		os.unlink(nginx_conf)

	reload_nginx()

def service(service, option):
	if os.path.basename(get_program(['systemctl']) or '') == 'systemctl' and is_running_systemd():
		exec_cmd("sudo {service_manager} {option} {service}".format(service_manager='systemctl', option=option, service=service))
	elif os.path.basename(get_program(['service']) or '') == 'service':
		exec_cmd("sudo {service_manager} {service} {option} ".format(service_manager='service', service=service, option=option))
	else:
		# look for 'service_manager' and 'service_manager_command' in environment
		service_manager = os.environ.get("BENCH_SERVICE_MANAGER")
		if service_manager:
			service_manager_command = (os.environ.get("BENCH_SERVICE_MANAGER_COMMAND")
				or "{service_manager} {option} {service}").format(service_manager=service_manager, service=service, option=option)
			exec_cmd(service_manager_command)

		else:
			raise Exception('No service manager found')

def get_supervisor_confdir():
	possiblities = ('/etc/supervisor/conf.d', '/etc/supervisor.d/', '/etc/supervisord/conf.d', '/etc/supervisord.d')
	for possiblity in possiblities:
		if os.path.exists(possiblity):
			return possiblity

def remove_default_nginx_configs():
	default_nginx_configs = ['/etc/nginx/conf.d/default.conf', '/etc/nginx/sites-enabled/default']

	for conf_file in default_nginx_configs:
		if os.path.exists(conf_file):
			os.unlink(conf_file)


def is_centos7():
	return os.path.exists('/etc/redhat-release') and get_cmd_output("cat /etc/redhat-release | sed 's/Linux\ //g' | cut -d' ' -f3 | cut -d. -f1").strip() == '7'

def is_running_systemd():
	with open('/proc/1/comm') as f:
		comm = f.read().strip()
	if comm == "init":
		return False
	elif comm == "systemd":
		return True
	return False

def reload_supervisor():
	supervisorctl = find_executable('supervisorctl')

	try:
		# first try reread/update
		exec_cmd('sudo {0} reread'.format(supervisorctl))
		exec_cmd('sudo {0} update'.format(supervisorctl))
		return
	except CommandFailedError:
		pass

	try:
		# something is wrong, so try reloading
		exec_cmd('sudo {0} reload'.format(supervisorctl))
		return
	except CommandFailedError:
		pass

	try:
		# then try restart for centos
		service('supervisord', 'restart')
		return
	except CommandFailedError:
		pass

	try:
		# else try restart for ubuntu / debian
		service('supervisor', 'restart')
		return
	except CommandFailedError:
		pass

def reload_nginx():
	try:
		subprocess.check_output(['sudo', find_executable('nginx'), '-t'])
	except:
		raise

	service('nginx', 'reload')



#! env python
import json
import os
import sys
import semantic_version
import git
import requests
import getpass
import re
from requests.auth import HTTPBasicAuth
import requests.exceptions
from time import sleep
from .config.common_site_config import get_config
import click

branches_to_update = {
	'develop': [],
	'hotfix': ['develop']
}

github_username = None
github_password = None

def release(bench_path, app, bump_type, from_branch='develop', to_branch='master',
		remote='upstream', owner='frappe', repo_name=None):

	config = get_config(bench_path)

	if not config.get('release_bench'):
		print('bench not configured to release')
		sys.exit(1)

	if config.get('branches_to_update'):
		branches_to_update.update(config.get('branches_to_update'))

	validate(bench_path, config)

	bump(bench_path, app, bump_type, from_branch=from_branch, to_branch=to_branch, owner=owner,
		repo_name=repo_name, remote=remote)

def validate(bench_path, config):
	global github_username, github_password

	github_username = config.get('github_username')
	github_password = config.get('github_password')

	if not github_username:
		github_username = click.prompt('Username', type=str)

	if not github_password:
		github_password = getpass.getpass()

	r = requests.get('https://api.github.com/user', auth=HTTPBasicAuth(github_username, github_password))
	r.raise_for_status()

def bump(bench_path, app, bump_type, from_branch, to_branch, remote, owner, repo_name=None):
	assert bump_type in ['minor', 'major', 'patch', 'stable', 'prerelease']

	repo_path = os.path.join(bench_path, 'apps', app)
	update_branches_and_check_for_changelog(repo_path, from_branch, to_branch, remote=remote)
	message = get_release_message(repo_path, from_branch=from_branch, to_branch=to_branch, remote=remote)

	if not message:
		print('No commits to release')
		return

	print()
	print(message)
	print()

	click.confirm('Do you want to continue?', abort=True)

	new_version = bump_repo(repo_path, bump_type, from_branch=from_branch, to_branch=to_branch)
	commit_changes(repo_path, new_version)
	tag_name = create_release(repo_path, new_version, from_branch=from_branch, to_branch=to_branch)
	push_release(repo_path, from_branch=from_branch, to_branch=to_branch, remote=remote)
	create_github_release(repo_path, tag_name, message, remote=remote, owner=owner, repo_name=repo_name)
	print('Released {tag} for {repo_path}'.format(tag=tag_name, repo_path=repo_path))

def update_branches_and_check_for_changelog(repo_path, from_branch='develop', to_branch='master', remote='upstream'):

	update_branch(repo_path, to_branch, remote=remote)
	update_branch(repo_path, from_branch, remote=remote)

	for branch in branches_to_update[from_branch]:
		update_branch(repo_path, branch, remote=remote)

	git.Repo(repo_path).git.checkout(from_branch)
	check_for_unmerged_changelog(repo_path)

def update_branch(repo_path, branch, remote):
	print("updating local branch of", repo_path, 'using', remote + '/' + branch)

	repo = git.Repo(repo_path)
	g = repo.git
	g.fetch(remote)
	g.checkout(branch)
	g.reset('--hard', remote+'/'+branch)

def check_for_unmerged_changelog(repo_path):
	current = os.path.join(repo_path, os.path.basename(repo_path), 'change_log', 'current')
	if os.path.exists(current) and [f for f in os.listdir(current) if f != "readme.md"]:
		raise Exception("Unmerged change log! in " + repo_path)

def get_release_message(repo_path, from_branch='develop', to_branch='master', remote='upstream'):
	print('getting release message for', repo_path, 'comparing', to_branch, '...', from_branch)

	repo = git.Repo(repo_path)
	g = repo.git
	log = g.log('{remote}/{to_branch}..{remote}/{from_branch}'.format(
		remote=remote, to_branch=to_branch, from_branch=from_branch), '--format=format:%s', '--no-merges')

	if log:
		return "* " + log.replace('\n', '\n* ')

def bump_repo(repo_path, bump_type, from_branch='develop', to_branch='master'):
	current_version = get_current_version(repo_path)
	new_version = get_bumped_version(current_version, bump_type)

	print('bumping version from', current_version, 'to', new_version)

	set_version(repo_path, new_version)
	return new_version

def get_current_version(repo_path):
	# TODO clean this up!
	filename = os.path.join(repo_path, os.path.basename(repo_path), '__init__.py')
	with open(filename) as f:
		contents = f.read()
		match = re.search(r"^(\s*%s\s*=\s*['\\\"])(.+?)(['\"])(?sm)" % '__version__',
				contents)
		return match.group(2)

def get_bumped_version(version, bump_type):
	v = semantic_version.Version(version)
	if bump_type == 'minor':
		v.minor += 1
		v.patch = 0
		v.prerelease = None

	elif bump_type == 'major':
		v.major += 1
		v.minor = 0
		v.patch = 0
		v.prerelease = None

	elif bump_type == 'patch':
		v.patch += 1
		v.prerelease = None

	elif bump_type == 'stable':
		# remove pre-release tag
		v.prerelease = None

	elif bump_type == 'prerelease':
		if v.prerelease == None:
			v.prerelease = ('beta',)

		elif len(v.prerelease)==1:
			v.prerelease[1] = '1'

		else:
			v.prerelease[1] = str(int(v.prerelease[1]) + 1)

	return str(v)

def set_version(repo_path, version):
	set_filename_version(os.path.join(repo_path, os.path.basename(repo_path),'__init__.py'), version, '__version__')

	# TODO fix this
	# set_setuppy_version(repo_path, version)
	# set_versionpy_version(repo_path, version)
	# set_hooks_version(repo_path, version)

# def set_setuppy_version(repo_path, version):
# 	set_filename_version(os.path.join(repo_path, 'setup.py'), version, 'version')
#
# def set_versionpy_version(repo_path, version):
# 	set_filename_version(os.path.join(repo_path, os.path.basename(repo_path),'__version__.py'), version, '__version__')
#
# def set_hooks_version(repo_path, version):
# 	set_filename_version(os.path.join(repo_path, os.path.basename(repo_path),'hooks.py'), version, 'app_version')

def set_filename_version(filename, version_number, pattern):
	changed = []

	def inject_version(match):
		before, old, after = match.groups()
		changed.append(True)
		return before + version_number + after

	with open(filename) as f:
		contents = re.sub(r"^(\s*%s\s*=\s*['\\\"])(.+?)(['\"])(?sm)" % pattern,
				inject_version, f.read())

	if not changed:
		raise Exception('Could not find %s in %s', pattern, filename)

	with open(filename, 'w') as f:
		f.write(contents)

def commit_changes(repo_path, new_version):
	print('committing version change to', repo_path)

	repo = git.Repo(repo_path)
	app_name = os.path.basename(repo_path)
	repo.index.add([os.path.join(app_name, '__init__.py')])
	repo.index.commit('bumped to version {}'.format(new_version))

def create_release(repo_path, new_version, from_branch='develop', to_branch='master'):
	print('creating release for version', new_version)
	repo = git.Repo(repo_path)
	g = repo.git
	g.checkout(to_branch)
	try:
		g.merge(from_branch, '--no-ff')
	except git.exc.GitCommandError as e:
		handle_merge_error(e, source=from_branch, target=to_branch)

	tag_name = 'v' + new_version
	repo.create_tag(tag_name, message='Release {}'.format(new_version))
	g.checkout(from_branch)

	try:
		g.merge(to_branch)
	except git.exc.GitCommandError as e:
		handle_merge_error(e, source=to_branch, target=from_branch)

	for branch in branches_to_update[from_branch]:
		print('merging master into', branch)
		g.checkout(branch)
		try:
			g.merge(to_branch)
		except git.exc.GitCommandError as e:
			handle_merge_error(e, source=to_branch, target=branch)

	return tag_name

def handle_merge_error(e, source, target):
	print('-'*80)
	print('Error when merging {source} into {target}'.format(source=source, target=target))
	print(e)
	print('You can open a new terminal, try to manually resolve the conflict/error and continue')
	print('-'*80)
	click.confirm('Have you manually resolved the error?', abort=True)

def push_release(repo_path, from_branch='develop', to_branch='master', remote='upstream'):
	print('pushing branches', to_branch, from_branch, 'of', repo_path)
	repo = git.Repo(repo_path)
	g = repo.git
	args = [
		'{to_branch}:{to_branch}'.format(to_branch=to_branch),
		'{from_branch}:{from_branch}'.format(from_branch=from_branch)
	]

	for branch in branches_to_update[from_branch]:
		print('pushing {0} branch of'.format(branch), repo_path)
		args.append('{branch}:{branch}'.format(branch=branch))

	args.append('--tags')

	print(g.push(remote, *args))

def create_github_release(repo_path, tag_name, message, remote='upstream', owner='frappe', repo_name=None,
		gh_username=None, gh_password=None):

	print('creating release on github')

	global github_username, github_password
	if not (gh_username and gh_password):
		if not (github_username and github_password):
			raise Exception("No credentials")
		gh_username = github_username
		gh_password = github_password

	repo_name = repo_name or os.path.basename(repo_path)
	data = {
		'tag_name': tag_name,
		'target_commitish': 'master',
		'name': 'Release ' + tag_name,
		'body': message,
		'draft': False,
		'prerelease': False
	}
	for i in range(3):
		try:
			r = requests.post('https://api.github.com/repos/{owner}/{repo_name}/releases'.format(
				owner=owner, repo_name=repo_name),
				auth=HTTPBasicAuth(gh_username, gh_password), data=json.dumps(data))
			r.raise_for_status()
			break
		except requests.exceptions.HTTPError:
			print('request failed, retrying....')
			sleep(3*i + 1)
			if i !=2:
				continue
			else:
				print(r.json())
				raise
	return r




import os, json
from bench.config.common_site_config import get_config, put_config, get_common_site_config

def execute(bench_path):
	# deprecate bench config
	bench_config_path = os.path.join(bench_path, 'config.json')
	if not os.path.exists(bench_config_path):
		return

	with open(bench_config_path, "r") as f:
		bench_config = json.loads(f.read())

	common_site_config = get_common_site_config(bench_path)
	common_site_config.update(bench_config)
	put_config(common_site_config, bench_path)

	# remove bench/config.json
	os.remove(bench_config_path)

	# change keys
	config = get_config(bench_path)
	changed = False
	for from_key, to_key, default in (
			("celery_broker", "redis_queue", "redis://localhost:6379"),
			("async_redis_server", "redis_socketio", "redis://localhost:12311"),
			("cache_redis_server", "redis_cache", "redis://localhost:11311")
		):
		if from_key in config:
			config[to_key] = config[from_key]
			del config[from_key]
			changed = True

		elif to_key not in config:
			config[to_key] = default
			changed = True

	if changed:
		put_config(config, bench_path)



import click
import os, sys, logging, json, pwd, subprocess
from bench.utils import is_root, PatchError, drop_privileges, get_env_cmd, get_cmd_output, get_frappe
from bench.app import get_apps
from bench.config.common_site_config import get_config
from bench.commands import bench_command

logger = logging.getLogger('bench')
from_command_line = False

def cli():
	global from_command_line
	from_command_line = True

	check_uid()
	change_dir()
	change_uid()

	if len(sys.argv) > 2 and sys.argv[1] == "frappe":
		return old_frappe_cli()

	elif len(sys.argv) > 1 and sys.argv[1] in get_frappe_commands():
		return frappe_cmd()

	elif len(sys.argv) > 1 and sys.argv[1] in ("--site", "--verbose", "--force", "--profile"):
		return frappe_cmd()

	elif len(sys.argv) > 1 and sys.argv[1]=="--help":
		print(click.Context(bench_command).get_help())
		print()
		print(get_frappe_help())
		return

	elif len(sys.argv) > 1 and sys.argv[1] in get_apps():
		return app_cmd()

	else:
		try:
			# NOTE: this is the main bench command
			bench_command()
		except PatchError:
			sys.exit(1)

def check_uid():
	if cmd_requires_root() and not is_root():
		print('superuser privileges required for this command')
		sys.exit(1)

def cmd_requires_root():
	if len(sys.argv) > 2 and sys.argv[2] in ('production', 'sudoers', 'lets-encrypt', 'fonts',
		'print', 'firewall', 'ssh-port', 'role'):
		return True
	if len(sys.argv) >= 2 and sys.argv[1] in ('patch', 'renew-lets-encrypt', 'disable-production',
		'install'):
		return True

def change_dir():
	if os.path.exists('config.json') or "init" in sys.argv:
		return
	dir_path_file = '/etc/frappe_bench_dir'
	if os.path.exists(dir_path_file):
		with open(dir_path_file) as f:
			dir_path = f.read().strip()
		if os.path.exists(dir_path):
			os.chdir(dir_path)

def change_uid():
	if is_root() and not cmd_requires_root():
		frappe_user = get_config(".").get('frappe_user')
		if frappe_user:
			drop_privileges(uid_name=frappe_user, gid_name=frappe_user)
			os.environ['HOME'] = pwd.getpwnam(frappe_user).pw_dir
		else:
			print('You should not run this command as root')
			sys.exit(1)

def old_frappe_cli(bench_path='.'):
	f = get_frappe(bench_path=bench_path)
	os.chdir(os.path.join(bench_path, 'sites'))
	os.execv(f, [f] + sys.argv[2:])

def app_cmd(bench_path='.'):
	f = get_env_cmd('python', bench_path=bench_path)
	os.chdir(os.path.join(bench_path, 'sites'))
	os.execv(f, [f] + ['-m', 'frappe.utils.bench_helper'] + sys.argv[1:])

def frappe_cmd(bench_path='.'):
	f = get_env_cmd('python', bench_path=bench_path)
	os.chdir(os.path.join(bench_path, 'sites'))
	os.execv(f, [f] + ['-m', 'frappe.utils.bench_helper', 'frappe'] + sys.argv[1:])

def get_frappe_commands(bench_path='.'):
	python = get_env_cmd('python', bench_path=bench_path)
	sites_path = os.path.join(bench_path, 'sites')
	if not os.path.exists(sites_path):
		return []
	try:
		return json.loads(get_cmd_output("{python} -m frappe.utils.bench_helper get-frappe-commands".format(python=python), cwd=sites_path))
	except subprocess.CalledProcessError:
		return []

def get_frappe_help(bench_path='.'):
	python = get_env_cmd('python', bench_path=bench_path)
	sites_path = os.path.join(bench_path, 'sites')
	if not os.path.exists(sites_path):
		return []
	try:
		out = get_cmd_output("{python} -m frappe.utils.bench_helper get-frappe-help".format(python=python), cwd=sites_path)
		return "Framework commands:\n" + out.split('Commands:')[1]
	except subprocess.CalledProcessError:
		return ""



# wget setup_frappe.py | python
import os, sys, subprocess, getpass, json, multiprocessing, shutil, platform
from distutils.spawn import find_executable

tmp_bench_repo = '/tmp/.bench'

def install_bench(args):
	check_distribution_compatibility()
	check_brew_installed()
	# pre-requisites for bench repo cloning
	install_package('curl')
	install_package('wget')

	success = run_os_command({
		'apt-get': [
			'sudo apt-get update',
			'sudo apt-get install -y git build-essential python-setuptools python-dev libffi-dev libssl-dev'
		],
		'yum': [
			'sudo yum groupinstall -y "Development tools"',
			'sudo yum install -y epel-release redhat-lsb-core git python-setuptools python-devel openssl-devel libffi-devel'
		],
		# epel-release is required to install redis, so installing it before the playbook-run.
		# redhat-lsb-core is required, so that ansible can set ansible_lsb variable
	})

	if not find_executable("git"):
		success = run_os_command({
			'brew': 'brew install git'
		})

	if not success:
		print('Could not install pre-requisites. Please check for errors or install them manually.')
		return

	# secure pip installation
	if find_executable('pip'):
		run_os_command({
			'yum': 'sudo pip install --upgrade setuptools pip',
			'apt-get': 'sudo pip install --upgrade setuptools pip',
			'brew': "sudo pip install --upgrade setuptools pip --user"
		})

	else:
		if not os.path.exists("get-pip.py"):
			run_os_command({
				'apt-get': 'wget https://bootstrap.pypa.io/get-pip.py',
				'yum': 'wget https://bootstrap.pypa.io/get-pip.py'
			})

		success = run_os_command({
			'apt-get': 'sudo python get-pip.py',
			'yum': 'sudo python get-pip.py',
		})

		if success:
			run_os_command({
				'pip': 'sudo pip install --upgrade pip setuptools',
			})

	# Restricting ansible version due to following bug in ansible 2.1
	# https://github.com/ansible/ansible-modules-core/issues/3752
	success = run_os_command({
		'pip': "sudo pip install ansible==2.3.1"
	})

	if not success:
		could_not_install('Ansible')

	# clone bench repo
	if not args.run_travis:
		clone_bench_repo(args)

	if not args.user:
		if args.production:
			args.user = 'frappe'

		elif 'SUDO_USER' in os.environ:
			args.user = os.environ['SUDO_USER']

		else:
			args.user = getpass.getuser()

	if args.user == 'root':
		raise Exception('Please run this script as a non-root user with sudo privileges, but without using sudo or pass --user=USER')

	# create user if not exists
	extra_vars = vars(args)
	extra_vars.update(frappe_user=args.user)

	if os.path.exists(tmp_bench_repo):
		repo_path = tmp_bench_repo

	else:
		repo_path = os.path.join(os.path.expanduser('~'), 'bench')

	extra_vars.update(repo_path=repo_path)
	run_playbook('develop/create_user.yml', extra_vars=extra_vars)

	extra_vars.update(get_passwords(args))
	if args.production:
		extra_vars.update(max_worker_connections=multiprocessing.cpu_count() * 1024)

	branch = 'master' if args.production else 'develop'
	extra_vars.update(branch=branch)
	
	bench_name = 'frappe-bench' if not args.bench_name else args.bench_name
	extra_vars.update(bench_name=bench_name)

	if args.develop:
		run_playbook('develop/install.yml', sudo=True, extra_vars=extra_vars)

	elif args.production:
		run_playbook('production/install.yml', sudo=True, extra_vars=extra_vars)

	if os.path.exists(tmp_bench_repo):
		shutil.rmtree(tmp_bench_repo)

def check_distribution_compatibility():
	supported_dists = {'ubuntu': [14, 15, 16], 'debian': [7, 8],
		'centos': [7], 'macos': [10.9, 10.10, 10.11, 10.12]}

	dist_name, dist_version = get_distribution_info()
	if dist_name in supported_dists:
		if float(dist_version) in supported_dists[dist_name]:
			return

	print("Sorry, the installer doesn't support {0} {1}. Aborting installation!".format(dist_name, dist_version))
	if dist_name in supported_dists:
		print("Install on {0} {1} instead".format(dist_name, supported_dists[dist_name][-1]))
	sys.exit(1)

def get_distribution_info():
	# return distribution name and major version
	if platform.system() == "Linux":
		current_dist = platform.dist()
		return current_dist[0].lower(), current_dist[1].rsplit('.')[0]
	elif platform.system() == "Darwin":
		current_dist = platform.mac_ver()
		return "macos", current_dist[0].rsplit('.', 1)[0]

def install_python27():
	version = (sys.version_info[0], sys.version_info[1])

	if version == (2, 7):
		return

	print('Installing Python 2.7')

	# install python 2.7
	success = run_os_command({
		'apt-get': 'sudo apt-get install -y python2.7',
		'yum': 'sudo yum install -y python27',
		'brew': 'brew install python'
	})

	if not success:
		could_not_install('Python 2.7')

	# replace current python with python2.7
	os.execvp('python2.7', ([] if is_sudo_user() else ['sudo']) + ['python2.7', __file__] + sys.argv[1:])

def install_package(package):
	package_exec = find_executable(package)

	if not package_exec:
		success = run_os_command({
			'apt-get': ['sudo apt-get install -y {0}'.format(package)],
			'yum': ['sudo yum install -y {0}'.format(package)]
		})
	else:
		return

	if not success:
		could_not_install(package)

def check_brew_installed():
	if 'Darwin' not in os.uname():
		return

	brew_exec = find_executable('brew')

	if not brew_exec:
		raise Exception('''
		Please install brew package manager before proceeding with bench setup. Please run following
		to install brew package manager on your machine,

		/usr/bin/ruby -e "$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/master/install)"
		''')

def clone_bench_repo(args):
	'''Clones the bench repository in the user folder'''
	if os.path.exists(tmp_bench_repo):
		return 0

	elif args.without_bench_setup:
		clone_path = os.path.join(os.path.expanduser('~'), 'bench')

	else:
		clone_path = tmp_bench_repo

	branch = args.bench_branch or 'master'
	repo_url = args.repo_url or 'https://github.com/frappe/bench'


	success = run_os_command(
		{'git': 'git clone {repo_url} {bench_repo} --depth 1 --branch {branch}'.format(
			repo_url=repo_url, bench_repo=clone_path, branch=branch)}
	)

	return success

def run_os_command(command_map):
	'''command_map is a dictionary of {'executable': command}. For ex. {'apt-get': 'sudo apt-get install -y python2.7'} '''
	success = True
	for executable, commands in list(command_map.items()):
		if find_executable(executable):
			if isinstance(commands, str):
				commands = [commands]

			for command in commands:
				returncode = subprocess.check_call(command, shell=True)
				success = success and ( returncode == 0 )

			break

	return success

def could_not_install(package):
	raise Exception('Could not install {0}. Please install it manually.'.format(package))

def is_sudo_user():
	return os.geteuid() == 0


def get_passwords(args):
	"""
	Returns a dict of passwords for further use
	and creates passwords.txt in the bench user's home directory
	"""

	ignore_prompt = args.run_travis or args.without_bench_setup
	mysql_root_password, admin_password = '', ''
	passwords_file_path = os.path.join(os.path.expanduser('~' + args.user), 'passwords.txt')

	if not ignore_prompt:
		# set passwords from existing passwords.txt
		if os.path.isfile(passwords_file_path):
			with open(passwords_file_path, 'r') as f:
				passwords = json.load(f)
				mysql_root_password, admin_password = passwords['mysql_root_password'], passwords['admin_password']

		# set passwords from cli args
		if args.mysql_root_password:
			mysql_root_password = args.mysql_root_password
		if args.admin_password:
			admin_password = args.admin_password

		# prompt for passwords
		pass_set = True
		while pass_set:
			# mysql root password
			if not mysql_root_password:
				mysql_root_password = getpass.unix_getpass(prompt='Please enter mysql root password: ')
				conf_mysql_passwd = getpass.unix_getpass(prompt='Re-enter mysql root password: ')

				if mysql_root_password != conf_mysql_passwd:
					mysql_root_password = ''
					continue

			# admin password
			if not admin_password:
				admin_password = getpass.unix_getpass(prompt='Please enter the default Administrator user password: ')
				conf_admin_passswd = getpass.unix_getpass(prompt='Re-enter Administrator password: ')

				if admin_password != conf_admin_passswd:
					admin_password = ''
					continue

			pass_set = False
	else:
		mysql_root_password = admin_password = 'travis'

	passwords = {
		'mysql_root_password': mysql_root_password,
		'admin_password': admin_password
	}

	if not ignore_prompt:
		with open(passwords_file_path, 'w') as f:
			json.dump(passwords, f, indent=1)

		print('Passwords saved at ~/passwords.txt')

	return passwords


def get_extra_vars_json(extra_args):
	# We need to pass production as extra_vars to the playbook to execute conditionals in the
	# playbook. Extra variables can passed as json or key=value pair. Here, we will use JSON.
	json_path = os.path.join('/tmp', 'extra_vars.json')
	extra_vars = dict(list(extra_args.items()))
	with open(json_path, mode='w') as j:
		json.dump(extra_vars, j, indent=1, sort_keys=True)

	return ('@' + json_path)

def run_playbook(playbook_name, sudo=False, extra_vars=None):
	args = ['ansible-playbook', '-c', 'local',  playbook_name]

	if extra_vars:
		args.extend(['-e', get_extra_vars_json(extra_vars)])

		if extra_vars.get('verbosity'):
			args.append('-vvvv')

	if sudo:
		user = extra_vars.get('user') or getpass.getuser()
		args.extend(['--become', '--become-user={0}'.format(user)])

	if os.path.exists(tmp_bench_repo):
		cwd = tmp_bench_repo
	else:
		cwd = os.path.join(os.path.expanduser('~'), 'bench')

	success = subprocess.check_call(args, cwd=os.path.join(cwd, 'playbooks'))
	return success

def parse_commandline_args():
	import argparse

	parser = argparse.ArgumentParser(description='Frappe Installer')

	# Arguments develop and production are mutually exclusive both can't be specified together.
	# Hence, we need to create a group for discouraging use of both options at the same time.
	args_group = parser.add_mutually_exclusive_group()

	args_group.add_argument('--develop', dest='develop', action='store_true', default=False,
		help='Install developer setup')

	args_group.add_argument('--production', dest='production', action='store_true',
		default=False, help='Setup Production environment for bench')

	parser.add_argument('--site', dest='site', action='store', default='site1.local',
		help='Specifiy name for your first ERPNext site')
	
	parser.add_argument('--without-site', dest='without_site', action='store_true',
		default=False)

	parser.add_argument('--verbose', dest='verbosity', action='store_true', default=False,
		help='Run the script in verbose mode')

	parser.add_argument('--user', dest='user', help='Install frappe-bench for this user')

	parser.add_argument('--bench-branch', dest='bench_branch', help='Clone a particular branch of bench repository')

	parser.add_argument('--repo-url', dest='repo_url', help='Clone bench from the given url')

	# To enable testing of script using Travis, this should skip the prompt
	parser.add_argument('--run-travis', dest='run_travis', action='store_true', default=False,
		help=argparse.SUPPRESS)

	parser.add_argument('--without-bench-setup', dest='without_bench_setup', action='store_true', default=False,
		help=argparse.SUPPRESS)
	
	# whether to overwrite an existing bench
	parser.add_argument('--overwrite', dest='overwrite', action='store_true', default=False,
		help='Whether to overwrite an existing bench')

	# set passwords
	parser.add_argument('--mysql-root-password', dest='mysql_root_password', help='Set mysql root password')
	parser.add_argument('--admin-password', dest='admin_password', help='Set admin password')
	parser.add_argument('--bench-name', dest='bench_name', help='Create bench with specified name. Default name is frappe-bench')

	args = parser.parse_args()

	return args

if __name__ == '__main__':
	try:
		import argparse
	except ImportError:
		# install python2.7
		install_python27()

	args = parse_commandline_args()

	install_bench(args)

	print('''Frappe/ERPNext has been successfully installed!''')



import os, importlib

def run(bench_path):
	source_patch_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'patches.txt')
	target_patch_file = os.path.join(os.path.abspath(bench_path), 'patches.txt')

	with open(source_patch_file, 'r') as f:
		patches = [p.strip() for p in f.read().splitlines()
			if p.strip() and not p.strip().startswith("#")]

	executed_patches = []
	if os.path.exists(target_patch_file):
		with open(target_patch_file, 'r') as f:
			executed_patches = f.read().splitlines()

	try:
		for patch in patches:
			if patch not in executed_patches:
				module = importlib.import_module(patch.split()[0])
				execute = getattr(module, 'execute')
				result = execute(bench_path)

				if result != False:
					executed_patches.append(patch)

	finally:
		with open(target_patch_file, 'w') as f:
			f.write('\n'.join(executed_patches))

			# end with an empty line
			f.write('\n')

def set_all_patches_executed(bench_path):
	source_patch_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'patches.txt')
	target_patch_file = os.path.join(os.path.abspath(bench_path), 'patches.txt')

	with open(target_patch_file, 'w') as tf:
		with open(source_patch_file, 'r') as sf:
			tf.write(sf.read())




from bench.tests import test_init
from bench.config.production_setup import setup_production, get_supervisor_confdir, disable_production
import bench.utils
import os
import getpass
import re
import unittest
import time

class TestSetupProduction(test_init.TestBenchInit):
	# setUp, tearDown and other tests are defiend in TestBenchInit

	def test_setup_production(self):
		self.test_multiple_benches()

		user = getpass.getuser()

		for bench_name in ("test-bench-1", "test-bench-2"):
			bench_path = os.path.join(os.path.abspath(self.benches_path), bench_name)
			setup_production(user, bench_path)
			self.assert_nginx_config(bench_name)
			self.assert_supervisor_config(bench_name)

		# test after start of both benches
		for bench_name in ("test-bench-1", "test-bench-2"):
			self.assert_supervisor_process(bench_name)

		self.assert_nginx_process()

		# sudoers
		bench.utils.setup_sudoers(user)
		self.assert_sudoers(user)

		for bench_name in ("test-bench-1", "test-bench-2"):
			bench_path = os.path.join(os.path.abspath(self.benches_path), bench_name)
			disable_production(bench_path)

	def test_disable_production(self):
		bench_name = 'test-disable-prod'
		self.test_init(bench_name, frappe_branch='master')

		user = getpass.getuser()

		bench_path = os.path.join(os.path.abspath(self.benches_path), bench_name)
		setup_production(user, bench_path)

		disable_production(bench_path)

		self.assert_nginx_link(bench_name)
		self.assert_supervisor_link(bench_name)
		self.assert_supervisor_process(bench_name=bench_name, disable_production=True)

	def assert_nginx_config(self, bench_name):
		conf_src = os.path.join(os.path.abspath(self.benches_path), bench_name, 'config', 'nginx.conf')
		conf_dest = "/etc/nginx/conf.d/{bench_name}.conf".format(bench_name=bench_name)

		self.assertTrue(os.path.exists(conf_src))
		self.assertTrue(os.path.exists(conf_dest))

		# symlink matches
		self.assertEqual(os.path.realpath(conf_dest), conf_src)

		# file content
		with open(conf_src, "r") as f:
			f = f.read().decode("utf-8")

			for key in (
					"upstream {bench_name}-frappe",
					"upstream {bench_name}-socketio-server"
				):
				self.assertTrue(key.format(bench_name=bench_name) in f)

	def assert_nginx_process(self):
		out = bench.utils.get_cmd_output("sudo nginx -t 2>&1")
		self.assertTrue("nginx: configuration file /etc/nginx/nginx.conf test is successful" in out)

	def assert_sudoers(self, user):
		sudoers_file = '/etc/sudoers.d/frappe'
		self.assertTrue(os.path.exists(sudoers_file))

		with open(sudoers_file, 'r') as f:
			sudoers = f.read().decode('utf-8')

		self.assertTrue('{user} ALL = (root) NOPASSWD: /usr/sbin/service nginx *'.format(user=user) in sudoers)
		self.assertTrue('{user} ALL = (root) NOPASSWD: /usr/bin/supervisorctl'.format(user=user) in sudoers)
		self.assertTrue('{user} ALL = (root) NOPASSWD: /usr/sbin/nginx'.format(user=user) in sudoers)

	def assert_supervisor_config(self, bench_name, use_rq=True):
		conf_src = os.path.join(os.path.abspath(self.benches_path), bench_name, 'config', 'supervisor.conf')

		supervisor_conf_dir = get_supervisor_confdir()
		conf_dest = "{supervisor_conf_dir}/{bench_name}.conf".format(supervisor_conf_dir=supervisor_conf_dir, bench_name=bench_name)

		self.assertTrue(os.path.exists(conf_src))
		self.assertTrue(os.path.exists(conf_dest))

		# symlink matches
		self.assertEqual(os.path.realpath(conf_dest), conf_src)

		# file content
		with open(conf_src, "r") as f:
			f = f.read().decode("utf-8")

			tests = [
				"program:{bench_name}-frappe-web",
				"program:{bench_name}-redis-cache",
				"program:{bench_name}-redis-queue",
				"program:{bench_name}-redis-socketio",
				"program:{bench_name}-node-socketio",
				"group:{bench_name}-web",
				"group:{bench_name}-workers",
				"group:{bench_name}-redis"
			]

			if use_rq:
				tests.extend([
					"program:{bench_name}-frappe-schedule",
					"program:{bench_name}-frappe-default-worker",
					"program:{bench_name}-frappe-short-worker",
					"program:{bench_name}-frappe-long-worker"
				])

			else:
				tests.extend([
					"program:{bench_name}-frappe-workerbeat",
					"program:{bench_name}-frappe-worker",
					"program:{bench_name}-frappe-longjob-worker",
					"program:{bench_name}-frappe-async-worker"
				])

			for key in tests:
				self.assertTrue(key.format(bench_name=bench_name) in f)

	def assert_supervisor_process(self, bench_name, use_rq=True, disable_production=False):
		out = bench.utils.get_cmd_output("sudo supervisorctl status")

		if "STARTING" in out:
			time.sleep(10)
			out = bench.utils.get_cmd_output("sudo supervisorctl status")

		tests = [
			"{bench_name}-web:{bench_name}-frappe-web[\s]+RUNNING",
			"{bench_name}-web:{bench_name}-node-socketio[\s]+RUNNING",
			"{bench_name}-redis:{bench_name}-redis-cache[\s]+RUNNING",
			"{bench_name}-redis:{bench_name}-redis-queue[\s]+RUNNING",
			"{bench_name}-redis:{bench_name}-redis-socketio[\s]+RUNNING"
		]

		if use_rq:
			tests.extend([
				"{bench_name}-workers:{bench_name}-frappe-schedule[\s]+RUNNING",
				"{bench_name}-workers:{bench_name}-frappe-default-worker-0[\s]+RUNNING",
				"{bench_name}-workers:{bench_name}-frappe-short-worker-0[\s]+RUNNING",
				"{bench_name}-workers:{bench_name}-frappe-long-worker-0[\s]+RUNNING"
			])

		else:
			tests.extend([
				"{bench_name}-workers:{bench_name}-frappe-workerbeat[\s]+RUNNING",
				"{bench_name}-workers:{bench_name}-frappe-worker[\s]+RUNNING",
				"{bench_name}-workers:{bench_name}-frappe-longjob-worker[\s]+RUNNING",
				"{bench_name}-workers:{bench_name}-frappe-async-worker[\s]+RUNNING"
			])

		for key in tests:
			if disable_production:
				self.assertFalse(re.search(key.format(bench_name=bench_name), out))
			else:
				self.assertTrue(re.search(key.format(bench_name=bench_name), out))

	def assert_nginx_link(self, bench_name):
		nginx_conf_name = '{bench_name}.conf'.format(bench_name=bench_name)
		nginx_conf_path = os.path.join('/etc/nginx/conf.d', nginx_conf_name)

		self.assertFalse(os.path.islink(nginx_conf_path))

	def assert_supervisor_link(self, bench_name):
		supervisor_conf_dir = get_supervisor_confdir()
		supervisor_conf_name = '{bench_name}.conf'.format(bench_name=bench_name)
		supervisor_conf_path = os.path.join(supervisor_conf_dir, supervisor_conf_name)

		self.assertFalse(os.path.islink(supervisor_conf_path))




import unittest
import json, os, shutil, subprocess
import bench
import bench.utils
import bench.app
import bench.config.common_site_config
import bench.cli

bench.cli.from_command_line = True

class TestBenchInit(unittest.TestCase):
	def setUp(self):
		self.benches_path = "."
		self.benches = []

	def tearDown(self):
		for bench_name in self.benches:
			bench_path = os.path.join(self.benches_path, bench_name)
			if os.path.exists(bench_path):
				shutil.rmtree(bench_path, ignore_errors=True)

	def test_init(self, bench_name="test-bench", **kwargs):
		self.init_bench(bench_name, **kwargs)

		self.assert_folders(bench_name)

		self.assert_virtual_env(bench_name)

		self.assert_common_site_config(bench_name, bench.config.common_site_config.default_config)

		self.assert_config(bench_name)

		self.assert_socketio(bench_name)

	def test_multiple_benches(self):
		# 1st bench
		self.test_init("test-bench-1")

		self.assert_common_site_config("test-bench-1", {
			"webserver_port": 8000,
			"socketio_port": 9000,
			"file_watcher_port": 6787,
			"redis_queue": "redis://localhost:11000",
			"redis_socketio": "redis://localhost:12000",
			"redis_cache": "redis://localhost:13000"
		})

		# 2nd bench
		self.test_init("test-bench-2")

		self.assert_common_site_config("test-bench-2", {
			"webserver_port": 8001,
			"socketio_port": 9001,
			"file_watcher_port": 6788,
			"redis_queue": "redis://localhost:11001",
			"redis_socketio": "redis://localhost:12001",
			"redis_cache": "redis://localhost:13001"
		})

	def test_new_site(self):
		self.init_bench('test-bench')
		self.new_site("test-site-1.dev")

	def new_site(self, site_name):
		new_site_cmd = ["bench", "new-site", site_name, "--admin-password", "admin"]

		# set in travis
		if os.environ.get("TRAVIS"):
			new_site_cmd.extend(["--mariadb-root-password", "travis"])

		subprocess.check_output(new_site_cmd, cwd=os.path.join(self.benches_path, "test-bench"))

		site_path = os.path.join(self.benches_path, "test-bench", "sites", site_name)

		self.assertTrue(os.path.exists(site_path))
		self.assertTrue(os.path.exists(os.path.join(site_path, "private", "backups")))
		self.assertTrue(os.path.exists(os.path.join(site_path, "private", "files")))
		self.assertTrue(os.path.exists(os.path.join(site_path, "public", "files")))

		site_config_path = os.path.join(site_path, "site_config.json")
		self.assertTrue(os.path.exists(site_config_path))
		with open(site_config_path, "r") as f:
			site_config = json.loads(f.read())

		for key in ("db_name", "db_password"):
			self.assertTrue(key in site_config)
			self.assertTrue(site_config[key])

	def test_get_app(self):
		site_name = "test-site-2.dev"
		self.init_bench('test-bench')

		self.new_site(site_name)
		bench_path = os.path.join(self.benches_path, "test-bench")

		bench.app.get_app("https://github.com/frappe/frappe-client", bench_path=bench_path)
		self.assertTrue(os.path.exists(os.path.join(bench_path, "apps", "frappeclient")))

	def test_install_app(self):
		site_name = "test-site-3.dev"
		self.init_bench('test-bench')

		self.new_site(site_name)
		bench_path = os.path.join(self.benches_path, "test-bench")

		# get app
		bench.app.get_app("https://github.com/frappe/erpnext", "develop", bench_path=bench_path)

		self.assertTrue(os.path.exists(os.path.join(bench_path, "apps", "erpnext")))

		# install app
		bench.app.install_app("erpnext", bench_path=bench_path)

		# install it to site
		subprocess.check_output(["bench", "--site", site_name, "install-app", "erpnext"], cwd=bench_path)

		out = subprocess.check_output(["bench", "--site", site_name, "list-apps"], cwd=bench_path).decode()
		self.assertTrue("erpnext" in out)


	def test_remove_app(self):
		self.init_bench('test-bench')

		bench_path = os.path.join(self.benches_path, "test-bench")

		# get app
		bench.app.get_app("https://github.com/frappe/erpnext", "develop", bench_path=bench_path)

		self.assertTrue(os.path.exists(os.path.join(bench_path, "apps", "erpnext")))

		# remove it
		bench.app.remove_app("erpnext", bench_path=bench_path)

		self.assertFalse(os.path.exists(os.path.join(bench_path, "apps", "erpnext")))


	def test_switch_to_branch(self):
		self.init_bench('test-bench')

		bench_path = os.path.join(self.benches_path, "test-bench")
		app_path = os.path.join(bench_path, "apps", "frappe")

		bench.app.switch_branch(branch="master", apps=["frappe"], bench_path=bench_path, check_upgrade=False)
		out = subprocess.check_output(['git', 'status'], cwd=app_path).decode()
		self.assertTrue("master" in out)

		# bring it back to develop!
		bench.app.switch_branch(branch="develop", apps=["frappe"], bench_path=bench_path, check_upgrade=False)
		out = subprocess.check_output(['git', 'status'], cwd=app_path).decode()
		self.assertTrue("develop" in out)

	def init_bench(self, bench_name, **kwargs):
		self.benches.append(bench_name)
		bench.utils.init(bench_name, **kwargs)

	def test_drop_site(self):
		self.init_bench('test-bench')
		# Check without archive_path given to drop-site command
		self.drop_site("test-drop-without-archive-path")

		# Check with archive_path given to drop-site command
		home = os.path.abspath(os.path.expanduser('~'))
		archived_sites_path = os.path.join(home, 'archived_sites')

		self.drop_site("test-drop-with-archive-path", archived_sites_path=archived_sites_path)

	def drop_site(self, site_name, archived_sites_path=None):
		self.new_site(site_name)

		drop_site_cmd = ['bench', 'drop-site', site_name]

		if archived_sites_path:
			drop_site_cmd.extend(['--archived-sites-path', archived_sites_path])

		if os.environ.get('TRAVIS'):
			drop_site_cmd.extend(['--root-password', 'travis'])

		bench_path = os.path.join(self.benches_path, 'test-bench')
		try:
			subprocess.check_output(drop_site_cmd, cwd=bench_path)
		except subprocess.CalledProcessError as err:
			print(err.output)

		if not archived_sites_path:
			archived_sites_path = os.path.join(bench_path, 'archived_sites')
			self.assertTrue(os.path.exists(archived_sites_path))
			self.assertTrue(os.path.exists(os.path.join(archived_sites_path, site_name)))

		else:
			self.assertTrue(os.path.exists(archived_sites_path))
			self.assertTrue(os.path.exists(os.path.join(archived_sites_path, site_name)))

	def assert_folders(self, bench_name):
		for folder in bench.utils.folders_in_bench:
			self.assert_exists(bench_name, folder)

		self.assert_exists(bench_name, "sites", "assets")
		self.assert_exists(bench_name, "apps", "frappe")
		self.assert_exists(bench_name, "apps", "frappe", "setup.py")

	def assert_virtual_env(self, bench_name):
		bench_path = os.path.abspath(bench_name)
		python = os.path.join(bench_path, "env", "bin", "python")
		python_path = bench.utils.get_cmd_output('{python} -c "import os; print os.path.dirname(os.__file__)"'.format(python=python))

		# part of bench's virtualenv
		self.assertTrue(python_path.startswith(bench_path))
		self.assert_exists(python_path)
		self.assert_exists(python_path, "site-packages")
		self.assert_exists(python_path, "site-packages", "IPython")
		self.assert_exists(python_path, "site-packages", "pip")

		site_packages = os.listdir(os.path.join(python_path, "site-packages"))
		
	def assert_config(self, bench_name):
		for config, search_key in (
			("redis_queue.conf", "redis_queue.rdb"),
			("redis_socketio.conf", "redis_socketio.rdb"),
			("redis_cache.conf", "redis_cache.rdb")):

			self.assert_exists(bench_name, "config", config)

			with open(os.path.join(bench_name, "config", config), "r") as f:
				f = f.read().decode("utf-8")
				self.assertTrue(search_key in f)

	def assert_socketio(self, bench_name):
		self.assert_exists(bench_name, "node_modules")
		self.assert_exists(bench_name, "node_modules", "socket.io")

	def assert_common_site_config(self, bench_name, expected_config):
		common_site_config_path = os.path.join(bench_name, 'sites', 'common_site_config.json')
		self.assertTrue(os.path.exists(common_site_config_path))

		config = self.load_json(common_site_config_path)

		for key, value in list(expected_config.items()):
			self.assertEqual(config.get(key), value)

	def assert_exists(self, *args):
		self.assertTrue(os.path.exists(os.path.join(*args)))

	def load_json(self, path):
		with open(path, "r") as f:
			return json.loads(f.read().decode("utf-8"))



import os
from .utils import (exec_cmd, get_frappe, check_git_for_shallow_clone, build_assets,
	restart_supervisor_processes, get_cmd_output, run_frappe_cmd, CommandFailedError)
from .config.common_site_config import get_config

import logging
import requests
import semantic_version
import json
import re
import subprocess
import bench
import sys
import shutil

logging.basicConfig(level="DEBUG")
logger = logging.getLogger(__name__)

class InvalidBranchException(Exception): pass
class InvalidRemoteException(Exception): pass

class MajorVersionUpgradeException(Exception):
	def __init__(self, message, upstream_version, local_version):
		super(MajorVersionUpgradeException, self).__init__(message)
		self.upstream_version = upstream_version
		self.local_version = local_version

def get_apps(bench_path='.'):
	try:
		with open(os.path.join(bench_path, 'sites', 'apps.txt')) as f:
			return f.read().strip().split('\n')
	except IOError:
		return []

def add_to_appstxt(app, bench_path='.'):
	apps = get_apps(bench_path=bench_path)
	if app not in apps:
		apps.append(app)
		return write_appstxt(apps, bench_path=bench_path)

def remove_from_appstxt(app, bench_path='.'):
	apps = get_apps(bench_path=bench_path)
	if app in apps:
		apps.remove(app)
		return write_appstxt(apps, bench_path=bench_path)

def write_appstxt(apps, bench_path='.'):
	with open(os.path.join(bench_path, 'sites', 'apps.txt'), 'w') as f:
		return f.write('\n'.join(apps))

def get_app(git_url, branch=None, bench_path='.', build_asset_files=True, verbose=False):
	#less verbose app install
	if '/' not in git_url:
		git_url = 'https://github.com/frappe/' + git_url
	#Gets repo name from URL
	repo_name = git_url.rsplit('/', 1)[1].rsplit('.', 1)[0]
	logger.info('getting app {}'.format(repo_name))
	shallow_clone = '--depth 1' if check_git_for_shallow_clone() else ''
	branch = '--branch {branch}'.format(branch=branch) if branch else ''

	exec_cmd("git clone {git_url} {branch} {shallow_clone} --origin upstream".format(
				git_url=git_url,
				shallow_clone=shallow_clone,
				branch=branch),
			cwd=os.path.join(bench_path, 'apps'))

	#Retrieves app name from setup.py
	app_path = os.path.join(bench_path, 'apps', repo_name, 'setup.py')
	with open(app_path, 'rb') as f:
		app_name = re.search(r'name\s*=\s*[\'"](.*)[\'"]', f.read().decode('utf-8')).group(1)
		if repo_name != app_name:
			apps_path = os.path.join(os.path.abspath(bench_path), 'apps')
			os.rename(os.path.join(apps_path, repo_name), os.path.join(apps_path, app_name))

	print('installing', app_name)
	install_app(app=app_name, bench_path=bench_path, verbose=verbose)

	if build_asset_files:
		build_assets(bench_path=bench_path)
	conf = get_config(bench_path=bench_path)
	if conf.get('restart_supervisor_on_update'):
		restart_supervisor_processes(bench_path=bench_path)

def new_app(app, bench_path='.'):
	# For backwards compatibility
	app = app.lower().replace(" ", "_").replace("-", "_")
	logger.info('creating new app {}'.format(app))
	apps = os.path.abspath(os.path.join(bench_path, 'apps'))
	bench.set_frappe_version(bench_path=bench_path)

	if bench.FRAPPE_VERSION == 4:
		exec_cmd("{frappe} --make_app {apps} {app}".format(frappe=get_frappe(bench_path=bench_path),
			apps=apps, app=app))
	else:
		run_frappe_cmd('make-app', apps, app, bench_path=bench_path)
	install_app(app, bench_path=bench_path)

def install_app(app, bench_path='.', verbose=False, no_cache=False):
	logger.info('installing {}'.format(app))
	# find_links = '--find-links={}'.format(conf.get('wheel_cache_dir')) if conf.get('wheel_cache_dir') else ''
	find_links = ''
	exec_cmd("{pip} install {quiet} {find_links} -e {app} {no_cache}".format(
				pip=os.path.join(bench_path, 'env', 'bin', 'pip'),
				quiet="-q" if not verbose else "",
				no_cache='--no-cache-dir' if not no_cache else '',
				app=os.path.join(bench_path, 'apps', app),
				find_links=find_links))
	add_to_appstxt(app, bench_path=bench_path)

def remove_app(app, bench_path='.'):
	if not app in get_apps(bench_path):
		print("No app named {0}".format(app))
		sys.exit(1)

	app_path = os.path.join(bench_path, 'apps', app)
	site_path = os.path.join(bench_path, 'sites')
	pip = os.path.join(bench_path, 'env', 'bin', 'pip')

	for site in os.listdir(site_path):
		req_file = os.path.join(site_path, site, 'site_config.json')
		if os.path.exists(req_file):
			out = subprocess.check_output(["bench", "--site", site, "list-apps"], cwd=bench_path).decode()
			if re.search(r'\b' + app + r'\b', out):
				print("Cannot remove, app is installed on site: {0}".format(site))
				sys.exit(1)

	exec_cmd(["{0} uninstall -y {1}".format(pip, app_path)])
	remove_from_appstxt(app, bench_path)
	shutil.rmtree(app_path)
	run_frappe_cmd("build", bench_path=bench_path)
	if get_config(bench_path).get('restart_supervisor_on_update'):
		restart_supervisor_processes(bench_path=bench_path)


def pull_all_apps(bench_path='.', reset=False):
	'''Check all apps if there no local changes, pull'''
	rebase = '--rebase' if get_config(bench_path).get('rebase_on_pull') else ''
	
	# chech for local changes
	if not reset:
		for app in get_apps(bench_path=bench_path):
			app_dir = get_repo_dir(app, bench_path=bench_path)
			if os.path.exists(os.path.join(app_dir, '.git')):
				out = subprocess.check_output(["git", "status"], cwd=app_dir).decode()
				if not re.search(r'nothing to commit, working (directory|tree) clean', out):
					print('''

Cannot proceed with update: You have local changes in app "{0}" that are not committed.

Here are your choices:

1. Merge the {0} app manually with "git pull" / "git pull --rebase" and fix conflicts.
1. Temporarily remove your changes with "git stash" or discard them completely
	with "bench update --reset" or for individual repositries "git reset --hard"
2. If your changes are helpful for others, send in a pull request via GitHub and
	wait for them to be merged in the core.'''.format(app))
					sys.exit(1)

	for app in get_apps(bench_path=bench_path):
		app_dir = get_repo_dir(app, bench_path=bench_path)
		if os.path.exists(os.path.join(app_dir, '.git')):
			remote = get_remote(app)
			logger.info('pulling {0}'.format(app))
			if reset:
				exec_cmd("git fetch --all", cwd=app_dir)
				exec_cmd("git reset --hard {remote}/{branch}".format(
					remote=remote, branch=get_current_branch(app,bench_path=bench_path)), cwd=app_dir)
			else:
				exec_cmd("git pull {rebase} {remote} {branch}".format(rebase=rebase,
					remote=remote, branch=get_current_branch(app, bench_path=bench_path)), cwd=app_dir)
			exec_cmd('find . -name "*.pyc" -delete', cwd=app_dir)


def is_version_upgrade(app='frappe', bench_path='.', branch=None):
	try:
		fetch_upstream(app, bench_path=bench_path)
	except CommandFailedError:
		raise InvalidRemoteException("No remote named upstream for {0}".format(app))

	upstream_version = get_upstream_version(app=app, branch=branch, bench_path=bench_path)

	if not upstream_version:
		raise InvalidBranchException("Specified branch of app {0} is not in upstream".format(app))

	local_version = get_major_version(get_current_version(app, bench_path=bench_path))
	upstream_version = get_major_version(upstream_version)

	if upstream_version - local_version > 0:
		return (True, local_version, upstream_version)

	return (False, local_version, upstream_version)

def get_current_frappe_version(bench_path='.'):
	try:
		return get_major_version(get_current_version('frappe', bench_path=bench_path))
	except IOError:
		return 0

def get_current_branch(app, bench_path='.'):
	repo_dir = get_repo_dir(app, bench_path=bench_path)
	return get_cmd_output("basename $(git symbolic-ref -q HEAD)", cwd=repo_dir)

def get_remote(app, bench_path='.'):
	repo_dir = get_repo_dir(app, bench_path=bench_path)
	contents = subprocess.check_output(['git', 'remote', '-v'], cwd=repo_dir,
									   stderr=subprocess.STDOUT).decode()
	if re.findall('upstream[\s]+', contents):
		remote = 'upstream'
	else:
		# get the first remote
		remote = contents.splitlines()[0].split()[0]

	return remote

def use_rq(bench_path):
	bench_path = os.path.abspath(bench_path)
	celery_app = os.path.join(bench_path, 'apps', 'frappe', 'frappe', 'celery_app.py')
	return not os.path.exists(celery_app)

def fetch_upstream(app, bench_path='.'):
	repo_dir = get_repo_dir(app, bench_path=bench_path)
	return subprocess.call(["git", "fetch", "upstream"], cwd=repo_dir)

def get_current_version(app, bench_path='.'):
	repo_dir = get_repo_dir(app, bench_path=bench_path)
	try:
		with open(os.path.join(repo_dir, os.path.basename(repo_dir), '__init__.py')) as f:
			return get_version_from_string(f.read())

	except AttributeError:
		# backward compatibility
		with open(os.path.join(repo_dir, 'setup.py')) as f:
			return get_version_from_string(f.read(), field='version')

def get_upstream_version(app, branch=None, bench_path='.'):
	repo_dir = get_repo_dir(app, bench_path=bench_path)
	if not branch:
		branch = get_current_branch(app, bench_path=bench_path)
	try:
		contents = subprocess.check_output(['git', 'show', 'upstream/{branch}:{app}/__init__.py'.format(branch=branch, app=app)], cwd=repo_dir, stderr=subprocess.STDOUT).decode()
	except subprocess.CalledProcessError as e:
		if "Invalid object" in e.output:
			return None
		else:
			raise
	return get_version_from_string(contents)

def get_upstream_url(app, bench_path='.'):
	repo_dir = get_repo_dir(app, bench_path=bench_path)
	return subprocess.check_output(['git', 'config', '--get', 'remote.upstream.url'], cwd=repo_dir).decode().strip()

def get_repo_dir(app, bench_path='.'):
	return os.path.join(bench_path, 'apps', app)

def switch_branch(branch, apps=None, bench_path='.', upgrade=False, check_upgrade=True):
	from .utils import update_requirements, update_npm_packages, backup_all_sites, patch_sites, build_assets, pre_upgrade, post_upgrade
	from . import utils
	apps_dir = os.path.join(bench_path, 'apps')
	version_upgrade = (False,)
	switched_apps = []

	if not apps:
		apps = [name for name in os.listdir(apps_dir)
			if os.path.isdir(os.path.join(apps_dir, name))]
		if branch=="v4.x.x":
			apps.append('shopping_cart')

	for app in apps:
		app_dir = os.path.join(apps_dir, app)
		if os.path.exists(app_dir):
			try:
				if check_upgrade:
					version_upgrade = is_version_upgrade(app=app, bench_path=bench_path, branch=branch)
					if version_upgrade[0] and not upgrade:
						raise MajorVersionUpgradeException("Switching to {0} will cause upgrade from {1} to {2}. Pass --upgrade to confirm".format(branch, version_upgrade[1], version_upgrade[2]), version_upgrade[1], version_upgrade[2])
				print("Switching for "+app)
				unshallow = "--unshallow" if os.path.exists(os.path.join(app_dir, ".git", "shallow")) else ""
				exec_cmd("git config --unset-all remote.upstream.fetch", cwd=app_dir)
				exec_cmd("git config --add remote.upstream.fetch '+refs/heads/*:refs/remotes/upstream/*'", cwd=app_dir)
				exec_cmd("git fetch upstream {unshallow}".format(unshallow=unshallow), cwd=app_dir)
				exec_cmd("git checkout {branch}".format(branch=branch), cwd=app_dir)
				exec_cmd("git merge upstream/{branch}".format(branch=branch), cwd=app_dir)
				switched_apps.append(app)
			except CommandFailedError:
				print("Error switching to branch {0} for {1}".format(branch, app))
			except InvalidRemoteException:
				print("Remote does not exist for app "+app)
			except InvalidBranchException:
				print("Branch {0} does not exist in Upstream for {1}".format(branch, app))

	if switched_apps:
		print("Successfully switched branches for:\n" + "\n".join(switched_apps))

	if version_upgrade[0] and upgrade:
		update_requirements()
		update_npm_packages()
		pre_upgrade(version_upgrade[1], version_upgrade[2])
		reload(utils)
		backup_all_sites()
		patch_sites()
		build_assets()
		post_upgrade(version_upgrade[1], version_upgrade[2])

def switch_to_branch(branch=None, apps=None, bench_path='.', upgrade=False):
	switch_branch(branch, apps=apps, bench_path=bench_path, upgrade=upgrade)

def switch_to_master(apps=None, bench_path='.', upgrade=True):
	switch_branch('master', apps=apps, bench_path=bench_path, upgrade=upgrade)

def switch_to_develop(apps=None, bench_path='.', upgrade=True):
	switch_branch('develop', apps=apps, bench_path=bench_path, upgrade=upgrade)

def get_version_from_string(contents, field='__version__'):
	match = re.search(r"^(\s*%s\s*=\s*['\\\"])(.+?)(['\"])(?sm)" % field,
			contents)
	return match.group(2)

def get_major_version(version):
	return semantic_version.Version(version).major

def install_apps_from_path(path, bench_path='.'):
	apps = get_apps_json(path)
	for app in apps:
		get_app(app['url'], branch=app.get('branch'), bench_path=bench_path, build_asset_files=False)

def get_apps_json(path):
	if path.startswith('http'):
		r = requests.get(path)
		return r.json()
	else:
		with open(path) as f:
			return json.load(f)



import os, json
from bench.utils import get_sites
from bench.config.nginx import make_nginx_conf
from collections import defaultdict

def get_site_config(site, bench_path='.'):
	config_path = os.path.join(bench_path, 'sites', site, 'site_config.json')
	if not os.path.exists(config_path):
		return {}
	with open(config_path) as f:
		return json.load(f)

def put_site_config(site, config, bench_path='.'):
	config_path = os.path.join(bench_path, 'sites', site, 'site_config.json')
	with open(config_path, 'w') as f:
		return json.dump(config, f, indent=1)

def update_site_config(site, new_config, bench_path='.'):
	config = get_site_config(site, bench_path=bench_path)
	config.update(new_config)
	put_site_config(site, config, bench_path=bench_path)

def set_nginx_port(site, port, bench_path='.', gen_config=True):
	set_site_config_nginx_property(site, {"nginx_port": port}, bench_path=bench_path, gen_config=gen_config)

def set_ssl_certificate(site, ssl_certificate, bench_path='.', gen_config=True):
	set_site_config_nginx_property(site, {"ssl_certificate": ssl_certificate}, bench_path=bench_path, gen_config=gen_config)

def set_ssl_certificate_key(site, ssl_certificate_key, bench_path='.', gen_config=True):
	set_site_config_nginx_property(site, {"ssl_certificate_key": ssl_certificate_key}, bench_path=bench_path, gen_config=gen_config)

def set_site_config_nginx_property(site, config, bench_path='.', gen_config=True):
	if site not in get_sites(bench_path=bench_path):
		raise Exception("No such site")
	update_site_config(site, config, bench_path=bench_path)
	if gen_config:
		make_nginx_conf(bench_path=bench_path)

def set_url_root(site, url_root, bench_path='.'):
	update_site_config(site, {"host_name": url_root}, bench_path=bench_path)

def add_domain(site, domain, ssl_certificate, ssl_certificate_key, bench_path='.'):
	domains = get_domains(site, bench_path)
	for d in domains:
		if (isinstance(d, dict) and d['domain']==domain) or d==domain:
			print("Domain {0} already exists".format(domain))
			return

	if ssl_certificate_key and ssl_certificate:
		domain = {
			'domain' : domain,
			'ssl_certificate': ssl_certificate,
			'ssl_certificate_key': ssl_certificate_key
		}

	domains.append(domain)
	update_site_config(site, { "domains": domains }, bench_path=bench_path)

def remove_domain(site, domain, bench_path='.'):
	domains = get_domains(site, bench_path)
	for i, d in enumerate(domains):
		if (isinstance(d, dict) and d['domain']==domain) or d==domain:
			domains.remove(d)
			break

	update_site_config(site, { 'domains': domains }, bench_path=bench_path)

def sync_domains(site, domains, bench_path='.'):
	"""Checks if there is a change in domains. If yes, updates the domains list."""
	changed = False
	existing_domains = get_domains_dict(get_domains(site, bench_path))
	new_domains = get_domains_dict(domains)

	if set(existing_domains.keys()) != set(new_domains.keys()):
		changed = True

	else:
		for d in list(existing_domains.values()):
			if d != new_domains.get(d['domain']):
				changed = True
				break

	if changed:
		# replace existing domains with this one
		update_site_config(site, { 'domains': domains }, bench_path='.')

	return changed

def get_domains(site, bench_path='.'):
	return get_site_config(site, bench_path=bench_path).get('domains') or []

def get_domains_dict(domains):
	domains_dict = defaultdict(dict)
	for d in domains:
		if isinstance(d, str):
			domains_dict[d] = { 'domain': d }

		elif isinstance(d, dict):
			domains_dict[d['domain']] = d

	return domains_dict



from .common_site_config import get_config
import re, os, subprocess, semantic_version
import bench

try:
	from urllib.parse import urlparse
except ImportError:
	from urlparse import urlparse

def generate_config(bench_path):
	config = get_config(bench_path)

	ports = {}
	for key in ('redis_cache', 'redis_queue', 'redis_socketio'):
		ports[key] = urlparse(config[key]).port

	write_redis_config(
		template_name='redis_queue.conf',
		context={
			"port": ports['redis_queue'],
			"bench_path": os.path.abspath(bench_path),
		},
		bench_path=bench_path
	)

	write_redis_config(
		template_name='redis_socketio.conf',
		context={
			"port": ports['redis_socketio'],
		},
		bench_path=bench_path
	)

	write_redis_config(
		template_name='redis_cache.conf',
		context={
			"maxmemory": config.get('cache_maxmemory', get_max_redis_memory()),
			"port": ports['redis_cache'],
			"redis_version": get_redis_version(),
		},
		bench_path=bench_path
	)

	# make pids folder
	pid_path = os.path.join(bench_path, "config", "pids")
	if not os.path.exists(pid_path):
		os.makedirs(pid_path)

def write_redis_config(template_name, context, bench_path):
	template = bench.env.get_template(template_name)

	if "pid_path" not in context:
		context["pid_path"] = os.path.abspath(os.path.join(bench_path, "config", "pids"))

	with open(os.path.join(bench_path, 'config', template_name), 'w') as f:
		f.write(template.render(**context))

def get_redis_version():
	version_string = subprocess.check_output('redis-server --version', shell=True).decode().strip()

	# extract version number from string
	version = re.findall("\d+\.\d+", version_string)
	if not version:
		return None

	version = semantic_version.Version(version[0], partial=True)
	return float('{major}.{minor}'.format(major=version.major, minor=version.minor))

def get_max_redis_memory():
	import psutil
	
	total_virtual_mem = psutil.virtual_memory().total/(pow(1024, 2))
	max_memory = int(total_virtual_mem * 0.05) # Max memory for redis is 5% of virtual memory

	if max_memory < 50:
		return 50
	else:
		return max_memory



import os, sys, shutil, subprocess, logging, itertools, requests, json, platform, select, pwd, grp, multiprocessing, hashlib
from distutils.spawn import find_executable
import bench
from bench import env
from six import iteritems


class PatchError(Exception):
	pass

class CommandFailedError(Exception):
	pass

logger = logging.getLogger(__name__)

folders_in_bench = ('apps', 'sites', 'config', 'logs', 'config/pids')

def get_frappe(bench_path='.'):
	frappe = get_env_cmd('frappe', bench_path=bench_path)
	if not os.path.exists(frappe):
		print('frappe app is not installed. Run the following command to install frappe')
		print('bench get-app https://github.com/frappe/frappe.git')
	return frappe

def get_env_cmd(cmd, bench_path='.'):
	return os.path.abspath(os.path.join(bench_path, 'env', 'bin', cmd))

def init(path, apps_path=None, no_procfile=False, no_backups=False,
		no_auto_update=False, frappe_path=None, frappe_branch=None, wheel_cache_dir=None,
		verbose=False, clone_from=None, skip_bench_mkdir=False, skip_redis_config_generation=False):
	from .app import get_app, install_apps_from_path
	from .config.common_site_config import make_config
	from .config import redis
	from .config.procfile import setup_procfile
	from bench.patches import set_all_patches_executed

	if(skip_bench_mkdir):
		pass
	else:
		if os.path.exists(path):
			print('Directory {} already exists!'.format(path))
			raise Exception("Site directory already exists")
		os.makedirs(path)

	for dirname in folders_in_bench:
		try:
			os.makedirs(os.path.join(path, dirname))
		except OSError as e:
			if e.errno != os.errno.EEXIST:
				pass

	setup_logging()

	setup_env(bench_path=path)

	make_config(path)

	if clone_from:
		clone_apps_from(bench_path=path, clone_from=clone_from)
	else:
		if not frappe_path:
			frappe_path = 'https://github.com/frappe/frappe.git'

		get_app(frappe_path, branch=frappe_branch, bench_path=path, build_asset_files=False, verbose=verbose)

		if apps_path:
			install_apps_from_path(apps_path, bench_path=path)


	bench.set_frappe_version(bench_path=path)
	if bench.FRAPPE_VERSION > 5:
		update_npm_packages(bench_path=path)

	set_all_patches_executed(bench_path=path)
	build_assets(bench_path=path)

	if not skip_redis_config_generation:
		redis.generate_config(path)

	if not no_procfile:
		setup_procfile(path)
	if not no_backups:
		setup_backups(bench_path=path)
	if not no_auto_update:
		setup_auto_update(bench_path=path)

def clone_apps_from(bench_path, clone_from):
	from .app import install_app
	print('Copying apps from {0}...'.format(clone_from))
	subprocess.check_output(['cp', '-R', os.path.join(clone_from, 'apps'), bench_path])

	print('Copying node_modules from {0}...'.format(clone_from))
	subprocess.check_output(['cp', '-R', os.path.join(clone_from, 'node_modules'), bench_path])

	def setup_app(app):
		# run git reset --hard in each branch, pull latest updates and install_app
		app_path = os.path.join(bench_path, 'apps', app)
		if os.path.exists(os.path.join(app_path, '.git')):
			print('Cleaning up {0}'.format(app))

			# remove .egg-ino
			subprocess.check_output(['rm', '-rf', app + '.egg-info'], cwd=app_path)

			remotes = subprocess.check_output(['git', 'remote'], cwd=app_path).decode().strip().split()
			if 'upstream' in remotes:
				remote = 'upstream'
			else:
				remote = remotes[0]
			branch = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD'], cwd=app_path).decode().strip()
			subprocess.check_output(['git', 'reset', '--hard'], cwd=app_path)
			subprocess.check_output(['git', 'pull', '--rebase', remote, branch], cwd=app_path)

			install_app(app, bench_path)

	with open(os.path.join(clone_from, 'sites', 'apps.txt'), 'r') as f:
		apps = f.read().splitlines()

	for app in apps:
		setup_app(app)

def exec_cmd(cmd, cwd='.'):
	from .cli import from_command_line

	async = False if from_command_line else True
	if async:
		stderr = stdout = subprocess.PIPE
	else:
		stderr = stdout = None

	logger.info(cmd)

	p = subprocess.Popen(cmd, cwd=cwd, shell=True, stdout=stdout, stderr=stderr, universal_newlines=True)

	if async:
		return_code = print_output(p)
	else:
		return_code = p.wait()

	if return_code > 0:
		raise CommandFailedError(cmd)

def setup_env(bench_path='.'):
	exec_cmd('virtualenv -q {} -p {}'.format('env', sys.executable), cwd=bench_path)
	exec_cmd('./env/bin/pip -q install --upgrade pip', cwd=bench_path)
	exec_cmd('./env/bin/pip -q install wheel', cwd=bench_path)
	# exec_cmd('./env/bin/pip -q install https://github.com/frappe/MySQLdb1/archive/MySQLdb-1.2.5-patched.tar.gz', cwd=bench_path)
	exec_cmd('./env/bin/pip -q install six', cwd=bench_path)
	exec_cmd('./env/bin/pip -q install -e git+https://github.com/frappe/python-pdfkit.git#egg=pdfkit', cwd=bench_path)

def setup_socketio(bench_path='.'):
	exec_cmd("npm install socket.io redis express superagent cookie babel-core less chokidar \
		babel-cli babel-preset-es2015 babel-preset-es2016 babel-preset-es2017 babel-preset-babili", cwd=bench_path)

def new_site(site, mariadb_root_password=None, admin_password=None, bench_path='.'):
	"""
	Creates a new site in the specified bench, default is current bench
	"""

	logger.info('creating new site {}'.format(site))

	# consider an existing passwords.txt file
	passwords_file_path = os.path.join(os.path.expanduser('~'), 'passwords.txt')
	if os.path.isfile(passwords_file_path):
		with open(passwords_file_path, 'r') as f:
			passwords = json.load(f)
			mariadb_root_password, admin_password = passwords['mysql_root_password'], passwords['admin_password']

	mysql_root_password_fragment = '--root_password {}'.format(mariadb_root_password) if mariadb_root_password else ''
	admin_password_fragment = '--admin_password {}'.format(admin_password) if admin_password else ''

	exec_cmd("{frappe} {site} --install {db_name} {mysql_root_password_fragment} {admin_password_fragment}".format(
				frappe=get_frappe(bench_path=bench_path),
				site=site,
				db_name=hashlib.sha1(site.encode()).hexdigest()[:10],
				mysql_root_password_fragment=mysql_root_password_fragment,
				admin_password_fragment=admin_password_fragment
			), cwd=os.path.join(bench_path, 'sites'))

	if len(get_sites(bench_path=bench_path)) == 1:
		exec_cmd("{frappe} --use {site}".format(frappe=get_frappe(bench_path=bench_path), site=site), cwd=os.path.join(bench_path, 'sites'))


def patch_sites(bench_path='.'):
	bench.set_frappe_version(bench_path=bench_path)

	try:
		if bench.FRAPPE_VERSION == 4:
			exec_cmd("{frappe} --latest all".format(frappe=get_frappe(bench_path=bench_path)), cwd=os.path.join(bench_path, 'sites'))
		else:
			run_frappe_cmd('--site', 'all', 'migrate', bench_path=bench_path)
	except subprocess.CalledProcessError:
		raise PatchError

def build_assets(bench_path='.'):
	bench.set_frappe_version(bench_path=bench_path)

	if bench.FRAPPE_VERSION == 4:
		exec_cmd("{frappe} --build".format(frappe=get_frappe(bench_path=bench_path)), cwd=os.path.join(bench_path, 'sites'))
	else:
		run_frappe_cmd('build', bench_path=bench_path)

def get_sites(bench_path='.'):
	sites_dir = os.path.join(bench_path, "sites")
	sites = [site for site in os.listdir(sites_dir)
		if os.path.isdir(os.path.join(sites_dir, site)) and site not in ('assets',)]
	return sites

def get_sites_dir(bench_path='.'):
	return os.path.abspath(os.path.join(bench_path, 'sites'))

def get_bench_dir(bench_path='.'):
	return os.path.abspath(bench_path)

def setup_auto_update(bench_path='.'):
	logger.info('setting up auto update')
	add_to_crontab('0 10 * * * cd {bench_dir} &&  {bench} update --auto >> {logfile} 2>&1'.format(bench_dir=get_bench_dir(bench_path=bench_path),
		bench=os.path.join(get_bench_dir(bench_path=bench_path), 'env', 'bin', 'bench'),
		logfile=os.path.join(get_bench_dir(bench_path=bench_path), 'logs', 'auto_update_log.log')))

def setup_backups(bench_path='.'):
	logger.info('setting up backups')
	bench_dir = get_bench_dir(bench_path=bench_path)
	bench.set_frappe_version(bench_path=bench_path)

	if bench.FRAPPE_VERSION == 4:
		backup_command = "cd {sites_dir} && {frappe} --backup all".format(frappe=get_frappe(bench_path=bench_path),)
	else:
		backup_command = "cd {bench_dir} && {bench} --site all backup".format(bench_dir=bench_dir, bench=sys.argv[0])

	add_to_crontab('0 */6 * * *  {backup_command} >> {logfile} 2>&1'.format(backup_command=backup_command,
		logfile=os.path.join(get_bench_dir(bench_path=bench_path), 'logs', 'backup.log')))

def add_to_crontab(line):
	current_crontab = read_crontab()
	if not line in current_crontab:
		cmd = ["crontab"]
		if platform.system() == 'FreeBSD':
			cmd = ["crontab", "-"]
		s = subprocess.Popen(cmd, stdin=subprocess.PIPE)
		s.stdin.write(current_crontab.encode())
		s.stdin.write((line + '\n').encode())
		s.stdin.close()

def read_crontab():
	try:
		return subprocess.check_output(["crontab", "-l"]).decode()
	except subprocess.CalledProcessError:
		return b""

def update_bench():
	logger.info('updating bench')

	# bench-repo folder
	cwd = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

	exec_cmd("git pull", cwd=cwd)

def setup_sudoers(user):
	if not os.path.exists('/etc/sudoers.d'):
		os.makedirs('/etc/sudoers.d')

		set_permissions = False
		if not os.path.exists('/etc/sudoers'):
			set_permissions = True

		with open('/etc/sudoers', 'a') as f:
			f.write('\n#includedir /etc/sudoers.d\n')

		if set_permissions:
			os.chmod('/etc/sudoers', 0o440)

	sudoers_file = '/etc/sudoers.d/frappe'

	template = env.get_template('frappe_sudoers')
	frappe_sudoers = template.render(**{
		'user': user,
		'service': find_executable('service'),
		'systemctl': find_executable('systemctl'),
		'supervisorctl': find_executable('supervisorctl'),
		'nginx': find_executable('nginx'),
		'bench': find_executable('bench')
	})

	with open(sudoers_file, 'w') as f:
		f.write(frappe_sudoers.encode('utf-8'))

	os.chmod(sudoers_file, 0o440)

def setup_logging(bench_path='.'):
	if os.path.exists(os.path.join(bench_path, 'logs')):
		logger = logging.getLogger('bench')
		log_file = os.path.join(bench_path, 'logs', 'bench.log')
		formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
		hdlr = logging.FileHandler(log_file)
		hdlr.setFormatter(formatter)
		logger.addHandler(hdlr)
		logger.setLevel(logging.DEBUG)

def get_program(programs):
	program = None
	for p in programs:
		program = find_executable(p)
		if program:
			break
	return program

def get_process_manager():
	return get_program(['foreman', 'forego', 'honcho'])

def start(no_dev=False, concurrency=None):
	program = get_process_manager()
	if not program:
		raise Exception("No process manager found")
	os.environ['PYTHONUNBUFFERED'] = "true"
	if not no_dev:
		os.environ['DEV_SERVER'] = "true"

	command = [program, 'start']
	if concurrency:
		command.extend(['-c', concurrency])

	os.execv(program, command)

def check_cmd(cmd, cwd='.'):
	try:
		subprocess.check_call(cmd, cwd=cwd, shell=True)
		return True
	except subprocess.CalledProcessError:
		return False

def get_git_version():
	'''returns git version from `git --version`
	extracts version number from string `get version 1.9.1` etc'''
	version = get_cmd_output("git --version").strip().split()[2]
	version = '.'.join(version.split('.')[0:2])
	return float(version)

def check_git_for_shallow_clone():
	from .config.common_site_config import get_config
	config = get_config('.')

	if config.get('release_bench'):
		return False

	if not config.get('shallow_clone'):
		return False

	git_version = get_git_version()
	if git_version > 1.9:
		return True
	return False

def get_cmd_output(cmd, cwd='.'):
	try:
		return subprocess.check_output(cmd, cwd=cwd, shell=True, stderr=open(os.devnull, 'wb')).decode().strip()
	except subprocess.CalledProcessError as e:
		if e.output:
			print(e.output)
		raise

def restart_supervisor_processes(bench_path='.', web_workers=False):
	from .config.common_site_config import get_config
	conf = get_config(bench_path=bench_path)
	bench_name = get_bench_name(bench_path)

	cmd = conf.get('supervisor_restart_cmd')
	if cmd:
		exec_cmd(cmd, cwd=bench_path)

	else:
		supervisor_status = subprocess.check_output(['sudo', 'supervisorctl', 'status'], cwd=bench_path).decode()

		if web_workers and '{bench_name}-web:'.format(bench_name=bench_name) in supervisor_status:
			group = '{bench_name}-web:	'.format(bench_name=bench_name)

		elif '{bench_name}-workers:'.format(bench_name=bench_name) in supervisor_status:
			group = '{bench_name}-workers: {bench_name}-web:'.format(bench_name=bench_name)

		# backward compatibility
		elif '{bench_name}-processes:'.format(bench_name=bench_name) in supervisor_status:
			group = '{bench_name}-processes:'.format(bench_name=bench_name)

		# backward compatibility
		else:
			group = 'frappe:'

		exec_cmd('sudo supervisorctl restart {group}'.format(group=group), cwd=bench_path)

def set_default_site(site, bench_path='.'):
	if not site in get_sites(bench_path=bench_path):
		raise Exception("Site not in bench")
	exec_cmd("{frappe} --use {site}".format(frappe=get_frappe(bench_path=bench_path), site=site),
			cwd=os.path.join(bench_path, 'sites'))

def update_requirements(bench_path='.'):
	print('Updating Python libraries...')
	pip = os.path.join(bench_path, 'env', 'bin', 'pip')

	# upgrade pip to latest
	exec_cmd("{pip} install --upgrade pip".format(pip=pip))

	apps_dir = os.path.join(bench_path, 'apps')

	# Update bench requirements
	bench_req_file = os.path.join(os.path.dirname(bench.__path__[0]), 'requirements.txt')
	install_requirements(pip, bench_req_file)

	for app in os.listdir(apps_dir):
		req_file = os.path.join(apps_dir, app, 'requirements.txt')
		install_requirements(pip, req_file)

def update_npm_packages(bench_path='.'):
	print('Updating node libraries...')
	apps_dir = os.path.join(bench_path, 'apps')
	package_json = {}

	for app in os.listdir(apps_dir):
		package_json_path = os.path.join(apps_dir, app, 'package.json')

		if os.path.exists(package_json_path):
			with open(package_json_path, "r") as f:
				app_package_json = json.loads(f.read())
				# package.json is usually a dict in a dict
				for key, value in iteritems(app_package_json):
					if not key in package_json:
						package_json[key] = value
					else:
						if isinstance(value, dict):
							package_json[key].update(value)
						elif isinstance(value, list):
							package_json[key].extend(value)
						else:
							package_json[key] = value

	if package_json is {}:
		with open(os.path.join(os.path.dirname(__file__), 'package.json'), 'r') as f:
			package_json = json.loads(f.read())

	with open(os.path.join(bench_path, 'package.json'), 'w') as f:
		f.write(json.dumps(package_json, indent=1, sort_keys=True))

	exec_cmd('npm install', cwd=bench_path)

def install_requirements(pip, req_file):
	if os.path.exists(req_file):
		exec_cmd("{pip} install -q -r {req_file}".format(pip=pip, req_file=req_file))

def backup_site(site, bench_path='.'):
	bench.set_frappe_version(bench_path=bench_path)

	if bench.FRAPPE_VERSION == 4:
		exec_cmd("{frappe} --backup {site}".format(frappe=get_frappe(bench_path=bench_path), site=site),
				cwd=os.path.join(bench_path, 'sites'))
	else:
		run_frappe_cmd('--site', site, 'backup', bench_path=bench_path)

def backup_all_sites(bench_path='.'):
	for site in get_sites(bench_path=bench_path):
		backup_site(site, bench_path=bench_path)

def is_root():
	if os.getuid() == 0:
		return True
	return False

def set_mariadb_host(host, bench_path='.'):
	update_common_site_config({'db_host': host}, bench_path=bench_path)

def update_common_site_config(ddict, bench_path='.'):
	update_json_file(os.path.join(bench_path, 'sites', 'common_site_config.json'), ddict)

def update_json_file(filename, ddict):
	if os.path.exists(filename):
		with open(filename, 'r') as f:
			content = json.load(f)

	else:
		content = {}

	content.update(ddict)
	with open(filename, 'w') as f:
		content = json.dump(content, f, indent=1, sort_keys=True)

def drop_privileges(uid_name='nobody', gid_name='nogroup'):
	# from http://stackoverflow.com/a/2699996
	if os.getuid() != 0:
		# We're not root so, like, whatever dude
		return

	# Get the uid/gid from the name
	running_uid = pwd.getpwnam(uid_name).pw_uid
	running_gid = grp.getgrnam(gid_name).gr_gid

	# Remove group privileges
	os.setgroups([])

	# Try setting the new uid/gid
	os.setgid(running_gid)
	os.setuid(running_uid)

	# Ensure a very conservative umask
	os.umask(0o22)

def fix_prod_setup_perms(bench_path='.', frappe_user=None):
	from .config.common_site_config import get_config
	files = [
		"logs/web.error.log",
		"logs/web.log",
		"logs/workerbeat.error.log",
		"logs/workerbeat.log",
		"logs/worker.error.log",
		"logs/worker.log",
		"config/nginx.conf",
		"config/supervisor.conf",
	]

	if not frappe_user:
		frappe_user = get_config(bench_path).get('frappe_user')

	if not frappe_user:
		print("frappe user not set")
		sys.exit(1)

	for path in files:
		if os.path.exists(path):
			uid = pwd.getpwnam(frappe_user).pw_uid
			gid = grp.getgrnam(frappe_user).gr_gid
			os.chown(path, uid, gid)

def fix_file_perms():
	for dir_path, dirs, files in os.walk('.'):
		for _dir in dirs:
			os.chmod(os.path.join(dir_path, _dir), 0o755)
		for _file in files:
			os.chmod(os.path.join(dir_path, _file), 0o644)
	bin_dir = './env/bin'
	if os.path.exists(bin_dir):
		for _file in os.listdir(bin_dir):
			if not _file.startswith('activate'):
				os.chmod(os.path.join(bin_dir, _file), 0o755)

def get_current_frappe_version(bench_path='.'):
	from .app import get_current_frappe_version as fv
	return fv(bench_path=bench_path)

def run_frappe_cmd(*args, **kwargs):
	from .cli import from_command_line

	bench_path = kwargs.get('bench_path', '.')
	f = get_env_cmd('python', bench_path=bench_path)
	sites_dir = os.path.join(bench_path, 'sites')

	async = False if from_command_line else True
	if async:
		stderr = stdout = subprocess.PIPE
	else:
		stderr = stdout = None

	p = subprocess.Popen((f, '-m', 'frappe.utils.bench_helper', 'frappe') + args,
		cwd=sites_dir, stdout=stdout, stderr=stderr)

	if async:
		return_code = print_output(p)
	else:
		return_code = p.wait()

	if return_code > 0:
		sys.exit(return_code)
		#raise CommandFailedError(args)

def get_frappe_cmd_output(*args, **kwargs):
	bench_path = kwargs.get('bench_path', '.')
	f = get_env_cmd('python', bench_path=bench_path)
	sites_dir = os.path.join(bench_path, 'sites')
	return subprocess.check_output((f, '-m', 'frappe.utils.bench_helper', 'frappe') + args, cwd=sites_dir).decode()

def validate_upgrade(from_ver, to_ver, bench_path='.'):
	if to_ver >= 6:
		if not find_executable('npm') and not (find_executable('node') or find_executable('nodejs')):
			raise Exception("Please install nodejs and npm")

def pre_upgrade(from_ver, to_ver, bench_path='.'):
	pip = os.path.join(bench_path, 'env', 'bin', 'pip')

	if from_ver <= 4 and to_ver >= 5:
		from .migrate_to_v5 import remove_shopping_cart
		apps = ('frappe', 'erpnext')
		remove_shopping_cart(bench_path=bench_path)

		for app in apps:
			cwd = os.path.abspath(os.path.join(bench_path, 'apps', app))
			if os.path.exists(cwd):
				exec_cmd("git clean -dxf", cwd=cwd)
				exec_cmd("{pip} install --upgrade -e {app}".format(pip=pip, app=cwd))

def post_upgrade(from_ver, to_ver, bench_path='.'):
	from .config.common_site_config import get_config
	from .config import redis
	from .config.supervisor import generate_supervisor_config
	from .config.nginx import make_nginx_conf
	conf = get_config(bench_path=bench_path)
	print("-"*80)
	print("Your bench was upgraded to version {0}".format(to_ver))

	if conf.get('restart_supervisor_on_update'):
		redis.generate_config(bench_path=bench_path)
		generate_supervisor_config(bench_path=bench_path)
		make_nginx_conf(bench_path=bench_path)

		if from_ver == 4 and to_ver == 5:
			setup_backups(bench_path=bench_path)

		if from_ver <= 5 and to_ver == 6:
			setup_socketio(bench_path=bench_path)

		print("As you have setup your bench for production, you will have to reload configuration for nginx and supervisor")
		print("To complete the migration, please run the following commands")
		print()
		print("sudo service nginx restart")
		print("sudo supervisorctl reload")

def update_translations_p(args):
	try:
		update_translations(*args)
	except requests.exceptions.HTTPError:
		print('Download failed for', args[0], args[1])

def download_translations_p():
	pool = multiprocessing.Pool(4)

	langs = get_langs()
	apps = ('frappe', 'erpnext')
	args = list(itertools.product(apps, langs))

	pool.map(update_translations_p, args)

def download_translations():
	langs = get_langs()
	apps = ('frappe', 'erpnext')
	for app, lang in itertools.product(apps, langs):
		update_translations(app, lang)

def get_langs():
	lang_file = 'apps/frappe/frappe/geo/languages.json'
	with open(lang_file) as f:
		langs = json.loads(f.read())
	return [d['code'] for d in langs]

def update_translations(app, lang):
	translations_dir = os.path.join('apps', app, app, 'translations')
	csv_file = os.path.join(translations_dir, lang + '.csv')
	url = "https://translate.erpnext.com/files/{}-{}.csv".format(app, lang)
	r = requests.get(url, stream=True)
	r.raise_for_status()

	with open(csv_file, 'wb') as f:
		for chunk in r.iter_content(chunk_size=1024):
			# filter out keep-alive new chunks
			if chunk:
				f.write(chunk)
				f.flush()

	print('downloaded for', app, lang)

def download_chart_of_accounts():
	charts_dir = os.path.join('apps', "erpnext", "erpnext", 'accounts', 'chart_of_accounts', "submitted")
	csv_file = os.path.join(translations_dir, lang + '.csv')
	url = "https://translate.erpnext.com/files/{}-{}.csv".format(app, lang)
	r = requests.get(url, stream=True)
	r.raise_for_status()

def print_output(p):
	while p.poll() is None:
		readx = select.select([p.stdout.fileno(), p.stderr.fileno()], [], [])[0]
		send_buffer = []
		for fd in readx:
			if fd == p.stdout.fileno():
				while 1:
					buf = p.stdout.read(1)
					if not len(buf):
						break
					if buf == '\r' or buf == '\n':
						send_buffer.append(buf)
						log_line(''.join(send_buffer), 'stdout')
						send_buffer = []
					else:
						send_buffer.append(buf)

			if fd == p.stderr.fileno():
				log_line(p.stderr.readline(), 'stderr')
	return p.poll()


def log_line(data, stream):
	if stream == 'stderr':
		return sys.stderr.write(data)
	return sys.stdout.write(data)

def get_output(*cmd):
	try:
		return subprocess.check_output(cmd).decode()
	except subprocess.CalledProcessError:
		return b""

def before_update(bench_path, requirements):
	validate_pillow_dependencies(bench_path, requirements)

def validate_pillow_dependencies(bench_path, requirements):
	if not requirements:
		return

	try:
		pip = os.path.join(bench_path, 'env', 'bin', 'pip')
		exec_cmd("{pip} install Pillow".format(pip=pip))

	except CommandFailedError:
		distro = platform.linux_distribution()
		distro_name = distro[0].lower()
		if "centos" in distro_name or "fedora" in distro_name:
			print("Please install these dependencies using the command:")
			print("sudo yum install libtiff-devel libjpeg-devel libzip-devel freetype-devel lcms2-devel libwebp-devel tcl-devel tk-devel")

			raise

		elif "ubuntu" in distro_name or "elementary os" in distro_name or "debian" in distro_name:
			print("Please install these dependencies using the command:")

			if "ubuntu" in distro_name and distro[1]=="12.04":
				print("sudo apt-get install -y libtiff4-dev libjpeg8-dev zlib1g-dev libfreetype6-dev liblcms2-dev libwebp-dev tcl8.5-dev tk8.5-dev python-tk")
			else:
				print("sudo apt-get install -y libtiff5-dev libjpeg8-dev zlib1g-dev libfreetype6-dev liblcms2-dev libwebp-dev tcl8.6-dev tk8.6-dev python-tk")

			raise

def get_bench_name(bench_path):
	return os.path.basename(os.path.abspath(bench_path))

def setup_fonts():
	fonts_path = os.path.join('/tmp', 'fonts')

	if os.path.exists('/etc/fonts_backup'):
		return

	exec_cmd("git clone https://github.com/frappe/fonts.git", cwd='/tmp')
	os.rename('/etc/fonts', '/etc/fonts_backup')
	os.rename('/usr/share/fonts', '/usr/share/fonts_backup')
	os.rename(os.path.join(fonts_path, 'etc_fonts'), '/etc/fonts')
	os.rename(os.path.join(fonts_path, 'usr_share_fonts'), '/usr/share/fonts')
	shutil.rmtree(fonts_path)
	exec_cmd("fc-cache -fv")

def set_git_remote_url(git_url, bench_path='.'):
	"Set app remote git url"
	app = git_url.rsplit('/', 1)[1].rsplit('.', 1)[0]

	if app not in bench.app.get_apps(bench_path):
		print("No app named {0}".format(app))
		sys.exit(1)

	app_dir = bench.app.get_repo_dir(app, bench_path=bench_path)
	if os.path.exists(os.path.join(app_dir, '.git')):
		exec_cmd("git remote set-url upstream {}".format(git_url), cwd=app_dir)

def run_playbook(playbook_name, extra_vars=None, tag=None):
	if not find_executable('ansible'):
		print("Ansible is needed to run this command, please install it using 'pip install ansible'")
		sys.exit(1)
	args = ['ansible-playbook', '-c', 'local', playbook_name]
	
	if extra_vars:
		args.extend(['-e', json.dumps(extra_vars)])
	
	if tag:
		args.extend(['-t', tag])
	
	subprocess.check_call(args, cwd=os.path.join(os.path.dirname(bench.__path__[0]), 'playbooks'))



import click
from bench.config.redis import generate_config

def execute(bench_path):
	click.confirm('\nThis update will replace ERPNext\'s Redis configuration files to fix a major security issue.\n'
		'If you don\'t know what this means, type Y ;)\n\n'
		'Do you want to continue?',
		abort=True)

	generate_config(bench_path)



import click, sys, json

@click.group()
def setup():
	"Setup bench"
	pass


@click.command('sudoers')
@click.argument('user')
def setup_sudoers(user):
	"Add commands to sudoers list for execution without password"
	from bench.utils import setup_sudoers
	setup_sudoers(user)


@click.command('nginx')
@click.option('--yes', help='Yes to regeneration of nginx config file', default=False, is_flag=True)
def setup_nginx(yes=False):
	"generate config for nginx"
	from bench.config.nginx import make_nginx_conf
	make_nginx_conf(bench_path=".", yes=yes)

@click.command('reload-nginx')
def reload_nginx():
	from bench.config.production_setup import reload_nginx
	reload_nginx()

@click.command('supervisor')
@click.option('--user')
@click.option('--yes', help='Yes to regeneration of supervisor config', is_flag=True, default=False)
def setup_supervisor(user=None, yes=False):
	"generate config for supervisor with an optional user argument"
	from bench.config.supervisor import generate_supervisor_config
	generate_supervisor_config(bench_path=".", user=user, yes=yes)

@click.command('redis')
def setup_redis():
	"generate config for redis cache"
	from bench.config.redis import generate_config
	generate_config('.')


@click.command('fonts')
def setup_fonts():
	"Add frappe fonts to system"
	from bench.utils import setup_fonts
	setup_fonts()


@click.command('production')
@click.argument('user')
@click.option('--yes', help='Yes to regeneration config', is_flag=True, default=False)
def setup_production(user, yes=False):
	"setup bench for production"
	from bench.config.production_setup import setup_production
	setup_production(user=user, yes=yes)


@click.command('auto-update')
def setup_auto_update():
	"Add cronjob for bench auto update"
	from bench.utils import setup_auto_update
	setup_auto_update()


@click.command('backups')
def setup_backups():
	"Add cronjob for bench backups"
	from bench.utils import setup_backups
	setup_backups()

@click.command('env')
def setup_env():
	"Setup virtualenv for bench"
	from bench.utils import setup_env
	setup_env()

@click.command('firewall')
@click.option('--ssh_port')
@click.option('--force')
def setup_firewall(ssh_port=None, force=False):
	"Setup firewall"
	from bench.utils import run_playbook

	if not force:
		click.confirm('Setting up the firewall will block all ports except 80, 443 and 22\n'
			'Do you want to continue?',
			abort=True)

	if not ssh_port:
		ssh_port = 22

	run_playbook('production/setup_firewall.yml', {"ssh_port": ssh_port})

@click.command('ssh-port')
@click.argument('port')
@click.option('--force')
def set_ssh_port(port, force=False):
	"Set SSH Port"
	from bench.utils import run_playbook

	if not force:
		click.confirm('This will change your SSH Port to {}\n'
			'Do you want to continue?'.format(port),
			abort=True)

	run_playbook('production/change_ssh_port.yml', {"ssh_port": port})

@click.command('lets-encrypt')
@click.argument('site')
@click.option('--custom-domain')
def setup_letsencrypt(site, custom_domain):
	"Setup lets-encrypt for site"
	from bench.config.lets_encrypt import setup_letsencrypt
	setup_letsencrypt(site, custom_domain, bench_path='.')


@click.command('procfile')
def setup_procfile():
	"Setup Procfile for bench start"
	from bench.config.procfile import setup_procfile
	setup_procfile('.')


@click.command('socketio')
def setup_socketio():
	"Setup node deps for socketio server"
	from bench.utils import setup_socketio
	setup_socketio()

@click.command('requirements')
def setup_requirements():
	"Setup python and node requirements"
	from bench.utils import update_requirements, update_npm_packages
	update_requirements()
	update_npm_packages()

@click.command('manager')
def setup_manager():
	"Setup bench_manager site and app"
	_new_site_ = "bench new-site bench-manager.local"
	_add_app_bench_manager_ = "bench get-app bench_manager https://github.com/frappe/bench_manager"
	_install_bench_manager_ = "bench --site bench-manager.local install-app bench_manager"
	c_list = [_new_site_, _add_app_bench_manager_, _install_bench_manager_]
	for c in c_list:
		out = subprocess.check_output(c.split())
		print(out)

@click.command('config')
def setup_config():
	"overwrite or make config.json"
	from bench.config.common_site_config import make_config
	make_config('.')


@click.command('add-domain')
@click.argument('domain')
@click.option('--site', prompt=True)
@click.option('--ssl-certificate', help="Absolute path to SSL Certificate")
@click.option('--ssl-certificate-key', help="Absolute path to SSL Certificate Key")
def add_domain(domain, site=None, ssl_certificate=None, ssl_certificate_key=None):
	"""Add custom domain to site"""
	from bench.config.site_config import add_domain

	if not site:
		print("Please specify site")
		sys.exit(1)

	add_domain(site, domain, ssl_certificate, ssl_certificate_key, bench_path='.')

@click.command('remove-domain')
@click.argument('domain')
@click.option('--site', prompt=True)
def remove_domain(domain, site=None):
	"""Remove custom domain from a site"""
	from bench.config.site_config import remove_domain

	if not site:
		print("Please specify site")
		sys.exit(1)

	remove_domain(site, domain, bench_path='.')

@click.command('sync-domains')
@click.option('--domain', multiple=True)
@click.option('--site', prompt=True)
def sync_domains(domain=None, site=None):
	from bench.config.site_config import sync_domains

	if not site:
		print("Please specify site")
		sys.exit(1)

	try:
		domains = list(map(str,domain))
	except Exception:
		print("Domains should be a json list of strings or dictionaries")
		sys.exit(1)

	changed = sync_domains(site, domains, bench_path='.')

	# if changed, success, else failure
	sys.exit(0 if changed else 1)

@click.command('role')
@click.argument('role')
@click.option('--admin_emails', default='')
@click.option('--mysql_root_password')
def setup_roles(role, **kwargs):
	"Install dependancies via roles"
	from bench.utils import run_playbook

	extra_vars = {"production": True}
	extra_vars.update(kwargs)

	if role:
		run_playbook('prerequisites/install_roles.yml', extra_vars=extra_vars, tag=role)
	else:
		run_playbook('prerequisites/install_roles.yml', extra_vars=extra_vars)


setup.add_command(setup_sudoers)
setup.add_command(setup_nginx)
setup.add_command(reload_nginx)
setup.add_command(setup_supervisor)
setup.add_command(setup_redis)
setup.add_command(setup_letsencrypt)
setup.add_command(setup_production)
setup.add_command(setup_auto_update)
setup.add_command(setup_backups)
setup.add_command(setup_env)
setup.add_command(setup_procfile)
setup.add_command(setup_socketio)
setup.add_command(setup_requirements)
setup.add_command(setup_config)
setup.add_command(setup_fonts)
setup.add_command(add_domain)
setup.add_command(remove_domain)
setup.add_command(sync_domains)
setup.add_command(setup_firewall)
setup.add_command(set_ssh_port)
setup.add_command(setup_roles)



import click

@click.command()
@click.argument('path')
@click.option('--apps_path', default=None, help="path to json files with apps to install after init")
@click.option('--frappe-path', default=None, help="path to frappe repo")
@click.option('--frappe-branch', default=None, help="path to frappe repo")
@click.option('--clone-from', default=None, help="copy repos from path")
@click.option('--no-procfile', is_flag=True, help="Pull changes in all the apps in bench")
@click.option('--no-backups',is_flag=True, help="Run migrations for all sites in the bench")
@click.option('--no-auto-update',is_flag=True, help="Build JS and CSS artifacts for the bench")
@click.option('--verbose',is_flag=True, help="Verbose output during install")
@click.option('--skip-bench-mkdir', is_flag=True, help="Skip mkdir frappe-bench")
@click.option('--skip-redis-config-generation', is_flag=True, help="Skip redis config generation if already specifying the common-site-config file")
def init(path, apps_path, frappe_path, frappe_branch, no_procfile, no_backups,
		no_auto_update, clone_from, verbose, skip_bench_mkdir, skip_redis_config_generation):
	"Create a new bench"
	from bench.utils import init
	init(path, apps_path=apps_path, no_procfile=no_procfile, no_backups=no_backups,
			no_auto_update=no_auto_update, frappe_path=frappe_path, frappe_branch=frappe_branch,
			verbose=verbose, clone_from=clone_from, skip_bench_mkdir=skip_bench_mkdir, skip_redis_config_generation=skip_redis_config_generation)
	click.echo('Bench {} initialized'.format(path))


@click.command('get-app')
@click.argument('name', nargs=-1) # Dummy argument for backward compatibility
@click.argument('git-url')
@click.option('--branch', default=None, help="branch to checkout")
def get_app(git_url, branch, name=None):
	"clone an app from the internet and set it up in your bench"
	from bench.app import get_app
	get_app(git_url, branch=branch)


@click.command('new-app')
@click.argument('app-name')
def new_app(app_name):
	"start a new app"
	from bench.app import new_app
	new_app(app_name)


@click.command('remove-app')
@click.argument('app-name')
def remove_app(app_name):
	"completely remove app from bench"
	from bench.app import remove_app
	remove_app(app_name)


@click.command('new-site')
@click.option('--mariadb-root-password', help="MariaDB root password")
@click.option('--admin-password', help="admin password to set for site")
@click.argument('site')
def new_site(site, mariadb_root_password=None, admin_password=None):
	"Create a new site in the bench"
	from bench.utils import new_site
	new_site(site, mariadb_root_password=mariadb_root_password, admin_password=admin_password)



import click, json
from bench.config.common_site_config import update_config

@click.group()
def config():
	"change bench configuration"
	pass

@click.command('auto_update')
@click.argument('state', type=click.Choice(['on', 'off']))
def config_auto_update(state):
	"Enable/Disable auto update for bench"
	state = True if state == 'on' else False
	update_config({'auto_update': state})


@click.command('restart_supervisor_on_update')
@click.argument('state', type=click.Choice(['on', 'off']))
def config_restart_supervisor_on_update(state):
	"Enable/Disable auto restart of supervisor processes"
	state = True if state == 'on' else False
	update_config({'restart_supervisor_on_update': state})


@click.command('update_bench_on_update')
@click.argument('state', type=click.Choice(['on', 'off']))
def config_update_bench_on_update(state):
	"Enable/Disable bench updates on running bench update"
	state = True if state == 'on' else False
	update_config({'update_bench_on_update': state})


@click.command('dns_multitenant')
@click.argument('state', type=click.Choice(['on', 'off']))
def config_dns_multitenant(state):
	"Enable/Disable bench updates on running bench update"
	state = True if state == 'on' else False
	update_config({'dns_multitenant': state})


@click.command('serve_default_site')
@click.argument('state', type=click.Choice(['on', 'off']))
def config_serve_default_site(state):
	"Configure nginx to serve the default site on port 80"
	state = True if state == 'on' else False
	update_config({'serve_default_site': state})


@click.command('rebase_on_pull')
@click.argument('state', type=click.Choice(['on', 'off']))
def config_rebase_on_pull(state):
	"Rebase repositories on pulling"
	state = True if state == 'on' else False
	update_config({'rebase_on_pull': state})


@click.command('http_timeout')
@click.argument('seconds', type=int)
def config_http_timeout(seconds):
	"set http timeout"
	update_config({'http_timeout': seconds})


@click.command('set-common-config')
@click.option('configs', '-c', '--config', multiple=True, type=(str, str))
def set_common_config(configs):
	import ast
	from bench.config.common_site_config import update_config

	common_site_config = {}
	for key, value in configs:
		if value in ("False", "True"):
			value = ast.literal_eval(value)

		elif "." in value:
			try:
				value = float(value)
			except ValueError:
				pass

		elif "{" in value or "[" in value:
			try:
				value = json.loads(value)
			except ValueError:
				pass

		else:
			try:
				value = int(value)
			except ValueError:
				pass

		common_site_config[key] = value

	update_config(common_site_config, bench_path='.')


@click.command('remove-common-config')
@click.argument('keys', nargs=-1)
def remove_common_config(keys):
	from bench.config.common_site_config import get_config, put_config
	common_site_config = get_config('.')
	for key in keys:
		if key in common_site_config:
			del common_site_config[key]

	put_config(common_site_config)


config.add_command(config_auto_update)
config.add_command(config_update_bench_on_update)
config.add_command(config_restart_supervisor_on_update)
config.add_command(config_dns_multitenant)
config.add_command(config_serve_default_site)
config.add_command(config_http_timeout)
config.add_command(set_common_config)
config.add_command(remove_common_config)



import os, json, click, random, string, hashlib
from bench.utils import get_sites, get_bench_name, exec_cmd

def make_nginx_conf(bench_path, yes=False):
	from bench import env
	from bench.config.common_site_config import get_config

	template = env.get_template('nginx.conf')
	bench_path = os.path.abspath(bench_path)
	sites_path = os.path.join(bench_path, "sites")

	config = get_config(bench_path)
	sites = prepare_sites(config, bench_path)
	bench_name = get_bench_name(bench_path)

	allow_rate_limiting = config.get('allow_rate_limiting', False)

	template_vars = {
		"sites_path": sites_path,
		"http_timeout": config.get("http_timeout"),
		"sites": sites,
		"webserver_port": config.get('webserver_port'),
		"socketio_port": config.get('socketio_port'),
		"bench_name": bench_name,
		"error_pages": get_error_pages(),
		"allow_rate_limiting": allow_rate_limiting,
		# for nginx map variable
		"random_string": "".join(random.choice(string.ascii_lowercase) for i in range(7))
	}

	if allow_rate_limiting:
		template_vars.update({
			'bench_name_hash': hashlib.sha256(bench_name).hexdigest()[:16],
			'limit_conn_shared_memory': get_limit_conn_shared_memory()
		})

	nginx_conf = template.render(**template_vars)

	conf_path = os.path.join(bench_path, "config", "nginx.conf")
	if not yes and os.path.exists(conf_path):
		click.confirm('nginx.conf already exists and this will overwrite it. Do you want to continue?',
			abort=True)

	with open(conf_path, "w") as f:
		f.write(nginx_conf)

def prepare_sites(config, bench_path):
	sites = {
		"that_use_port": [],
		"that_use_dns": [],
		"that_use_ssl": [],
		"that_use_wildcard_ssl": []
	}

	domain_map = {}
	ports_in_use = {}

	dns_multitenant = config.get('dns_multitenant')

	shared_port_exception_found = False
	sites_configs = get_sites_with_config(bench_path=bench_path)


	# preload all preset site ports to avoid conflicts

	if not dns_multitenant:
		for site in sites_configs:
			if site.get("port"):
				if not site["port"] in ports_in_use:
					ports_in_use[site["port"]] = []
				ports_in_use[site["port"]].append(site["name"])

	for site in sites_configs:
		if dns_multitenant:
			domain = site.get('domain')

			if domain:
				# when site's folder name is different than domain name
				domain_map[domain] = site['name']

			site_name = domain or site['name']

			if site.get('wildcard'):
				sites["that_use_wildcard_ssl"].append(site_name)

				if not sites.get('wildcard_ssl_certificate'):
					sites["wildcard_ssl_certificate"] = site['ssl_certificate']
					sites["wildcard_ssl_certificate_key"] = site['ssl_certificate_key']

			elif site.get("ssl_certificate") and site.get("ssl_certificate_key"):
				sites["that_use_ssl"].append(site)

			else:
				sites["that_use_dns"].append(site_name)

		else:
			if not site.get("port"):
				site["port"] = 80
				if site["port"] in ports_in_use:
					site["port"] = 8001
				while site["port"] in ports_in_use:
					site["port"] += 1

#			if site["port"] in ports_in_use:
#				raise Exception("Port {0} is being used by another site {1}".format(site["port"], ports_in_use[site["port"]]))

			if site["port"] in ports_in_use and not site["name"] in ports_in_use[site["port"]]:
				shared_port_exception_found = True
				ports_in_use[site["port"]].append(site["name"])
			else:
				ports_in_use[site["port"]] = []
				ports_in_use[site["port"]].append(site["name"])

			sites["that_use_port"].append(site)


	if not dns_multitenant and shared_port_exception_found:
		message = "Port conflicts found:"
		port_conflict_index = 0
		for port_number in ports_in_use:
			if len(ports_in_use[port_number]) > 1:
				port_conflict_index += 1
				message += "\n{0} - Port {1} is shared among sites:".format(port_conflict_index,port_number)
				for site_name in ports_in_use[port_number]:
					message += " {0}".format(site_name)
		raise Exception(message)

	if not dns_multitenant:
		message = "Port configuration list:"
		port_config_index = 0
		for site in sites_configs:
			port_config_index += 1
			message += "\n\nSite {0} assigned port: {1}".format(site["name"], site["port"])

		print(message)


	sites['domain_map'] = domain_map

	return sites

def get_sites_with_config(bench_path):
	from bench.config.common_site_config import get_config
	from bench.config.site_config import get_site_config

	sites = get_sites(bench_path=bench_path)
	dns_multitenant = get_config(bench_path).get('dns_multitenant')

	ret = []
	for site in sites:
		site_config = get_site_config(site, bench_path=bench_path)
		ret.append({
			"name": site,
			"port": site_config.get('nginx_port'),
			"ssl_certificate": site_config.get('ssl_certificate'),
			"ssl_certificate_key": site_config.get('ssl_certificate_key')
		})

		if dns_multitenant and site_config.get('domains'):
			for domain in site_config.get('domains'):
				# domain can be a string or a dict with 'domain', 'ssl_certificate', 'ssl_certificate_key'
				if isinstance(domain, str) or isinstance(domain, unicode):
					domain = { 'domain': domain }

				domain['name'] = site
				ret.append(domain)

	use_wildcard_certificate(bench_path, ret)

	return ret

def use_wildcard_certificate(bench_path, ret):
	'''
		stored in common_site_config.json as:
	    "wildcard": {
			"domain": "*.erpnext.com",
			"ssl_certificate": "/path/to/erpnext.com.cert",
			"ssl_certificate_key": "/path/to/erpnext.com.key"
		}
	'''
	from bench.config.common_site_config import get_config
	config = get_config(bench_path=bench_path)
	wildcard = config.get('wildcard')

	if not wildcard:
		return

	domain = wildcard['domain']
	ssl_certificate = wildcard['ssl_certificate']
	ssl_certificate_key = wildcard['ssl_certificate_key']

	if domain.startswith('*.'):
		domain = domain[1:]
	else:
		domain = '.' + domain

	for site in ret:
		if site.get('ssl_certificate'):
			continue

		if (site.get('domain') or site['name']).endswith(domain):
			# example: ends with .erpnext.com
			site['ssl_certificate'] = ssl_certificate
			site['ssl_certificate_key'] = ssl_certificate_key
			site['wildcard'] = 1

def get_error_pages():
	import bench
	bench_app_path = os.path.abspath(bench.__path__[0])
	templates = os.path.join(bench_app_path, 'config', 'templates')

	return {
		502: os.path.join(templates, '502.html')
	}

def get_limit_conn_shared_memory():
	"""Allocate 2 percent of total virtual memory as shared memory for nginx limit_conn_zone"""
	import psutil
	total_vm = (psutil.virtual_memory().total) / (1024 * 1024) # in MB

	return int(0.02 * total_vm)



import bench, os, click, errno
from bench.utils import exec_cmd, CommandFailedError
from bench.config.site_config import update_site_config, remove_domain, get_domains
from bench.config.nginx import make_nginx_conf
from bench.config.production_setup import service
from bench.config.common_site_config import get_config
from crontab import CronTab

try:
	from urllib.request import urlretrieve
except ImportError:
	from urllib import urlretrieve

def setup_letsencrypt(site, custom_domain, bench_path):

	site_path = os.path.join(bench_path, "sites", site, "site_config.json")
	if not os.path.exists(os.path.dirname(site_path)):
		print("No site named "+site)
		return

	if custom_domain:
		domains = get_domains(site, bench_path)
		for d in domains:
			if (isinstance(d, dict) and d['domain']==custom_domain):
				print("SSL for Domain {0} already exists".format(custom_domain))
				return

		if not custom_domain in domains:
			print("No custom domain named {0} set for site".format(custom_domain))
			return

	click.confirm('Running this will stop the nginx service temporarily causing your sites to go offline\n'
		'Do you want to continue?',
		abort=True)

	if not get_config(bench_path).get("dns_multitenant"):
		print("You cannot setup SSL without DNS Multitenancy")
		return

	create_config(site, custom_domain)
	run_certbot_and_setup_ssl(site, custom_domain, bench_path)
	setup_crontab()


def create_config(site, custom_domain):
	config = bench.env.get_template('letsencrypt.cfg').render(domain=custom_domain or site)
	config_path = '/etc/letsencrypt/configs/{site}.cfg'.format(site=custom_domain or site)
	create_dir_if_missing(config_path)

	with open(config_path, 'w') as f:
		f.write(config)


def run_certbot_and_setup_ssl(site, custom_domain, bench_path):
	service('nginx', 'stop')
	get_certbot()

	try:
		exec_cmd("{path} --config /etc/letsencrypt/configs/{site}.cfg certonly".format(path=get_certbot_path(), site=custom_domain or site))
	except CommandFailedError:
		service('nginx', 'start')
		print("There was a problem trying to setup SSL for your site")
		return

	ssl_path = "/etc/letsencrypt/live/{site}/".format(site=custom_domain or site)
	ssl_config = { "ssl_certificate": os.path.join(ssl_path, "fullchain.pem"),
					"ssl_certificate_key": os.path.join(ssl_path, "privkey.pem") }

	if custom_domain:
		remove_domain(site, custom_domain, bench_path)
		domains = get_domains(site, bench_path)
		ssl_config['domain'] = custom_domain
		domains.append(ssl_config)
		update_site_config(site, { "domains": domains }, bench_path=bench_path)
	else:
		update_site_config(site, ssl_config, bench_path=bench_path)

	make_nginx_conf(bench_path)
	service('nginx', 'start')


def setup_crontab():
	job_command = 'sudo service nginx stop && /opt/certbot-auto renew && sudo service nginx start'
	system_crontab = CronTab(tabfile='/etc/crontab', user=True)
	if job_command not in str(system_crontab):
		job  = system_crontab.new(command=job_command, comment="Renew lets-encrypt every month")
		job.every().month()
		job.enable()
		system_crontab.write()


def create_dir_if_missing(path):
	if not os.path.exists(os.path.dirname(path)):
		os.makedirs(os.path.dirname(path))


def get_certbot():
	certbot_path = get_certbot_path()
	create_dir_if_missing(certbot_path)

	if not os.path.isfile(certbot_path):
		urlretrieve ("https://dl.eff.org/certbot-auto", certbot_path)
		os.chmod(certbot_path, 0o744)


def get_certbot_path():
	return "/opt/certbot-auto"


def renew_certs():
	click.confirm('Running this will stop the nginx service temporarily causing your sites to go offline\n'
		'Do you want to continue?',
		abort=True)

	service('nginx', 'stop')
	exec_cmd("{path} renew".format(path=get_certbot_path()))
	service('nginx', 'start')



#! env python
import os
import git
import click
from .config.common_site_config import get_config

github_username = None
github_password = None

def prepare_staging(bench_path, app, remote='upstream'):
	from .release import get_release_message
	validate(bench_path)

	repo_path = os.path.join(bench_path, 'apps', app)
	update_branches(repo_path, remote)
	message = get_release_message(repo_path, from_branch='develop', to_branch='staging', remote=remote)

	if not message:
		print('No commits to release')
		return

	print()
	print(message)
	print()

	click.confirm('Do you want to continue?', abort=True)

	create_staging(repo_path)
	push_commits(repo_path)

def validate(bench_path):
	from .release import validate

	config = get_config(bench_path)
	validate(bench_path, config)

def update_branches(repo_path, remote):
	from .release import update_branch
	update_branch(repo_path, 'staging', remote)
	update_branch(repo_path, 'develop', remote)

	git.Repo(repo_path).git.checkout('develop')

def create_staging(repo_path, from_branch='develop'):
	from .release import handle_merge_error

	print('creating staging from', from_branch)
	repo = git.Repo(repo_path)
	g = repo.git
	g.checkout('staging')
	try:
		g.merge(from_branch, '--no-ff')
	except git.exc.GitCommandError as e:
		handle_merge_error(e, source=from_branch, target='staging')
	
	g.checkout(from_branch)
	try:
		g.merge('staging')
	except git.exc.GitCommandError as e:
		handle_merge_error(e, source='staging', target=from_branch)
	
def push_commits(repo_path, remote='upstream'):
	print('pushing staging branch of', repo_path)

	repo = git.Repo(repo_path)
	g = repo.git

	args = [
		'develop:develop',
		'staging:staging'
	]

	print(g.push(remote, *args))



import click, subprocess, sys
from semantic_version import Version
from distutils.spawn import find_executable

def execute(bench_path):
	expected_node_ver = Version('5.0.0')
	node_exec = find_executable('node') or find_executable('nodejs')


	if node_exec:
		result = subprocess.check_output([node_exec, '-v']).decode()
	else:
		click.echo('''
		No node executable was found on your machine.
		Please install latest node version before running "bench update". For installation instructions
		please refer "Debian and Ubuntu based Linux distributions" section or "Enterprise Linux and
		Fedora" section depending upon your OS on the following link,
		"https://nodejs.org/en/download/package-manager/"
		''')
		sys.exit(1)

	node_ver = Version(result.rstrip('\n').lstrip('v'))

	if node_ver < expected_node_ver:
		click.echo('''
		Please update node to latest version before running "bench update".
		Please install latest node version before running "bench update". For installation instructions
		please refer "Debian and Ubuntu based Linux distributions" section or "Enterprise Linux and
		Fedora" section depending upon your OS on the following link,
		"https://nodejs.org/en/download/package-manager/"
		''')
		sys.exit(1)



# module  - bench
# imports - third-party imports
from jinja2 import Environment, PackageLoader

# imports - module imports
from bench.__attr__ import __version__, __release__

env = Environment(loader=PackageLoader('bench.config'))

FRAPPE_VERSION = None

def set_frappe_version(bench_path='.'):
	from .app import get_current_frappe_version
	global FRAPPE_VERSION
	if not FRAPPE_VERSION:
		FRAPPE_VERSION = get_current_frappe_version(bench_path=bench_path)



import click
import sys, os
from bench.config.common_site_config import get_config
from bench.app import pull_all_apps, is_version_upgrade
from bench.utils import (update_bench, validate_upgrade, pre_upgrade, post_upgrade, before_update,
	update_requirements, update_npm_packages, backup_all_sites, patch_sites, build_assets, restart_supervisor_processes)
from bench import patches

#TODO: Not DRY
@click.command('update')
@click.option('--pull', is_flag=True, help="Pull changes in all the apps in bench")
@click.option('--patch',is_flag=True, help="Run migrations for all sites in the bench")
@click.option('--build',is_flag=True, help="Build JS and CSS artifacts for the bench")
@click.option('--bench',is_flag=True, help="Update bench")
@click.option('--requirements',is_flag=True, help="Update requirements")
@click.option('--restart-supervisor',is_flag=True, help="restart supervisor processes after update")
@click.option('--auto',is_flag=True)
@click.option('--no-backup',is_flag=True)
@click.option('--force',is_flag=True)
@click.option('--reset', is_flag=True, help="Hard resets git branch's to their new states overriding any changes and overriding rebase on pull")
def update(pull=False, patch=False, build=False, bench=False, auto=False, restart_supervisor=False, requirements=False, no_backup=False, force=False, reset=False):
	"Update bench"

	if not (pull or patch or build or bench or requirements):
		pull, patch, build, bench, requirements = True, True, True, True, True

	if auto:
		sys.exit(1)

	patches.run(bench_path='.')
	conf = get_config(".")

	if bench and conf.get('update_bench_on_update'):
		update_bench()
		restart_update({
				'pull': pull,
				'patch': patch,
				'build': build,
				'requirements': requirements,
				'no-backup': no_backup,
				'restart-supervisor': restart_supervisor,
				'reset':reset
		})

	if conf.get('release_bench'):
		print('Release bench, cannot update')
		sys.exit(1)
	
	version_upgrade = is_version_upgrade()
	
	if version_upgrade[0]:
		print()
		print()
		print("This update will cause a major version change in Frappe/ERPNext from {0} to {1}.".format(*version_upgrade[1:]))
		print("This would take significant time to migrate and might break custom apps.")
		click.confirm('Do you want to continue?', abort=True)

	_update(pull, patch, build, bench, auto, restart_supervisor, requirements, no_backup, force=force, reset=reset)

def _update(pull=False, patch=False, build=False, update_bench=False, auto=False, restart_supervisor=False,
		requirements=False, no_backup=False, bench_path='.', force=False, reset=False):
	conf = get_config(bench_path=bench_path)
	version_upgrade = is_version_upgrade(bench_path=bench_path)

	if version_upgrade[0] or (not version_upgrade[0] and force):
		validate_upgrade(version_upgrade[1], version_upgrade[2], bench_path=bench_path)

	before_update(bench_path=bench_path, requirements=requirements)

	if pull:
		pull_all_apps(bench_path=bench_path, reset=reset)

	if requirements:
		update_requirements(bench_path=bench_path)
		update_npm_packages(bench_path=bench_path)

	if version_upgrade[0] or (not version_upgrade[0] and force):
		pre_upgrade(version_upgrade[1], version_upgrade[2], bench_path=bench_path)
		import bench.utils, bench.app
		print('Reloading bench...')
		reload(bench.utils)
		reload(bench.app)

	if patch:
		if not no_backup:
			print('Backing up sites...')
			backup_all_sites(bench_path=bench_path)

		print('Patching sites...')
		patch_sites(bench_path=bench_path)
	if build:
		build_assets(bench_path=bench_path)
	if version_upgrade[0] or (not version_upgrade[0] and force):
		post_upgrade(version_upgrade[1], version_upgrade[2], bench_path=bench_path)
	if restart_supervisor or conf.get('restart_supervisor_on_update'):
		restart_supervisor_processes(bench_path=bench_path)

	print("_"*80)
	print("Bench: Deployment tool for Frappe and ERPNext (https://erpnext.org).")
	print("Open source depends on your contributions, so please contribute bug reports, patches, fixes or cash and be a part of the community")
	print()

@click.command('retry-upgrade')
@click.option('--version', default=5)
def retry_upgrade(version):
	pull_all_apps()
	patch_sites()
	build_assets()
	post_upgrade(version-1, version)

def restart_update(kwargs):
	args = ['--'+k for k, v in list(kwargs.items()) if v]
	os.execv(sys.argv[0], sys.argv[:2] + args)

@click.command('switch-to-branch')
@click.argument('branch')
@click.argument('apps', nargs=-1)
@click.option('--upgrade',is_flag=True)
def switch_to_branch(branch, apps, upgrade=False):
	"Switch all apps to specified branch, or specify apps separated by space"
	from bench.app import switch_to_branch
	switch_to_branch(branch=branch, apps=list(apps), upgrade=upgrade)
	print('Switched to ' + branch)
	print('Please run `bench update --patch` to be safe from any differences in database schema')

@click.command('switch-to-master')
def switch_to_master():
	"Switch frappe and erpnext to master branch"
	from bench.app import switch_to_master
	switch_to_master(apps=['frappe', 'erpnext'])
	print()
	print('Switched to master')
	print('Please run `bench update --patch` to be safe from any differences in database schema')

@click.command('switch-to-develop')
def switch_to_develop(upgrade=False):
	"Switch frappe and erpnext to develop branch"
	from bench.app import switch_to_develop
	switch_to_develop(apps=['frappe', 'erpnext'])
	print()
	print('Switched to develop')
	print('Please run `bench update --patch` to be safe from any differences in database schema')



import click
import sys, os, copy


@click.command('start')
@click.option('--no-dev', is_flag=True, default=False)
@click.option('--concurrency', '-c', type=str)
def start(no_dev, concurrency):
	"Start Frappe development processes"
	from bench.utils import start
	start(no_dev=no_dev, concurrency=concurrency)


@click.command('restart')
@click.option('--web', is_flag=True, default=False)
def restart(web):
	"Restart supervisor processes"
	from bench.utils import restart_supervisor_processes
	restart_supervisor_processes(bench_path='.', web_workers=web)


@click.command('set-nginx-port')
@click.argument('site')
@click.argument('port', type=int)
def set_nginx_port(site, port):
	"Set nginx port for site"
	from bench.config.site_config import set_nginx_port
	set_nginx_port(site, port)


@click.command('set-ssl-certificate')
@click.argument('site')
@click.argument('ssl-certificate-path')
def set_ssl_certificate(site, ssl_certificate_path):
	"Set ssl certificate path for site"
	from bench.config.site_config import set_ssl_certificate
	set_ssl_certificate(site, ssl_certificate_path)


@click.command('set-ssl-key')
@click.argument('site')
@click.argument('ssl-certificate-key-path')
def set_ssl_certificate_key(site, ssl_certificate_key_path):
	"Set ssl certificate private key path for site"
	from bench.config.site_config import set_ssl_certificate_key
	set_ssl_certificate_key(site, ssl_certificate_key_path)


@click.command('set-url-root')
@click.argument('site')
@click.argument('url-root')
def set_url_root(site, url_root):
	"Set url root for site"
	from bench.config.site_config import set_url_root
	set_url_root(site, url_root)


@click.command('set-mariadb-host')
@click.argument('host')
def set_mariadb_host(host):
	"Set MariaDB host for bench"
	from bench.utils import set_mariadb_host
	set_mariadb_host(host)


@click.command('set-default-site')
@click.argument('site')
def set_default_site(site):
	"Set default site for bench"
	from bench.utils import set_default_site
	set_default_site(site)


@click.command('download-translations')
def download_translations():
	"Download latest translations"
	from bench.utils import download_translations_p
	download_translations_p()

@click.command('renew-lets-encrypt')
def renew_lets_encrypt():
	"Renew Let's Encrypt certificate"
	from bench.config.lets_encrypt import renew_certs
	renew_certs()

@click.command()
def shell(bench_path='.'):
	if not os.environ.get('SHELL'):
		print("Cannot get shell")
		sys.exit(1)
	if not os.path.exists('sites'):
		print("sites dir doesn't exist")
		sys.exit(1)
	env = copy.copy(os.environ)
	env['PS1'] = '(' + os.path.basename(os.path.dirname(os.path.abspath(__file__))) + ')' + env.get('PS1', '')
	env['PATH'] = os.path.dirname(os.path.abspath(os.path.join('env','bin')) + ':' + env['PATH'])
	os.chdir('sites')
	os.execve(env['SHELL'], [env['SHELL']], env)


@click.command('backup')
@click.argument('site')
def backup_site(site):
	"backup site"
	from bench.utils import get_sites, backup_site
	if not site in get_sites(bench_path='.'):
		print('site not found')
		sys.exit(1)
	backup_site(site, bench_path='.')


@click.command('backup-all-sites')
def backup_all_sites():
	"backup all sites"
	from bench.utils import backup_all_sites
	backup_all_sites(bench_path='.')


@click.command('release')
@click.argument('app')
@click.argument('bump-type', type=click.Choice(['major', 'minor', 'patch', 'stable', 'prerelease']))
@click.option('--from-branch', default='develop')
@click.option('--to-branch', default='master')
@click.option('--remote', default='upstream')
@click.option('--owner', default='frappe')
@click.option('--repo-name')
def release(app, bump_type, from_branch, to_branch, owner, repo_name, remote):
	"Release app (internal to the Frappe team)"
	from bench.release import release
	release(bench_path='.', app=app, bump_type=bump_type, from_branch=from_branch, to_branch=to_branch,
		remote=remote, owner=owner, repo_name=repo_name)

@click.command('prepare-staging')
@click.argument('app')
def prepare_staging(app):
	"""Prepare staging branch from develop branch"""
	from bench.prepare_staging import prepare_staging
	prepare_staging(bench_path='.', app=app)
	

@click.command('disable-production')
def disable_production():
	"""Disables production environment for the bench."""
	from bench.config.production_setup import disable_production
	disable_production(bench_path='.')


@click.command('src')
def bench_src():
	"""Prints bench source folder path, which can be used as: cd `bench src` """
	import bench
	print(os.path.dirname(bench.__path__[0]))



import os, multiprocessing, getpass, json

try:
	from urllib.parse import urlparse
except ImportError:
	from urlparse import urlparse

default_config = {
	'restart_supervisor_on_update': False,
	'auto_update': False,
	'serve_default_site': True,
	'rebase_on_pull': False,
	'update_bench_on_update': True,
	'frappe_user': getpass.getuser(),
	'shallow_clone': True,
	'background_workers': 1
}

def make_config(bench_path):
	make_pid_folder(bench_path)
	bench_config = get_config(bench_path)
	bench_config.update(default_config)
	bench_config.update(get_gunicorn_workers())
	update_config_for_frappe(bench_config, bench_path)

	put_config(bench_config, bench_path)

def get_config(bench_path):
	return get_common_site_config(bench_path)

def get_common_site_config(bench_path):
	config_path = get_config_path(bench_path)
	if not os.path.exists(config_path):
		return {}
	with open(config_path, 'r') as f:
		return json.load(f)

def put_config(config, bench_path='.'):
	config_path = get_config_path(bench_path)
	with open(config_path, 'w') as f:
		return json.dump(config, f, indent=1, sort_keys=True)

def update_config(new_config, bench_path='.'):
	config = get_config(bench_path=bench_path)
	config.update(new_config)
	put_config(config, bench_path=bench_path)

def get_config_path(bench_path):
	return os.path.join(bench_path, 'sites', 'common_site_config.json')

def get_gunicorn_workers():
	'''This function will return the maximum workers that can be started depending upon
	number of cpu's present on the machine'''
	return {
		"gunicorn_workers": multiprocessing.cpu_count()
	}

def update_config_for_frappe(config, bench_path):
	ports = make_ports(bench_path)

	for key in ('redis_cache', 'redis_queue', 'redis_socketio'):
		if key not in config:
			config[key] = "redis://localhost:{0}".format(ports[key])

	for key in ('webserver_port', 'socketio_port', 'file_watcher_port'):
		if key not in config:
			config[key] = ports[key]


	# TODO Optionally we need to add the host or domain name in case dns_multitenant is false

def make_ports(bench_path):
	benches_path = os.path.dirname(os.path.abspath(bench_path))

	default_ports = {
		"webserver_port": 8000,
		"socketio_port": 9000,
		"file_watcher_port": 6787,
		"redis_queue": 11000,
		"redis_socketio": 12000,
		"redis_cache": 13000
	}

	# collect all existing ports
	existing_ports = {}
	for folder in os.listdir(benches_path):
		bench_path = os.path.join(benches_path, folder)
		if os.path.isdir(bench_path):
			bench_config = get_config(bench_path)
			for key in list(default_ports.keys()):
				value = bench_config.get(key)

				# extract port from redis url
				if value and (key in ('redis_cache', 'redis_queue', 'redis_socketio')):
					value = urlparse(value).port

				if value:
					existing_ports.setdefault(key, []).append(value)

	# new port value = max of existing port value + 1
	ports = {}
	for key, value in list(default_ports.items()):
		existing_value = existing_ports.get(key, [])
		if existing_value:
			value = max(existing_value) + 1

		ports[key] = value

	return ports

def make_pid_folder(bench_path):
	pids_path = os.path.join(bench_path, 'config', 'pids')
	if not os.path.exists(pids_path):
		os.makedirs(pids_path)



#!/usr/bin/env python

# imports - standard imports
import sys, os, shutil
from distutils.core          import Command
from distutils.command.clean import clean as Clean

# imports - third-party imports
from setuptools import setup, find_packages

# imports - module imports
from package    import package

class CleanCommand(Clean):
	def run(self):
		Clean.run(self)

		basedir = os.path.abspath(os.path.dirname(__file__))

		for relpath in ['build', '.cache', '.coverage', 'dist', '{name}.egg-info'.format(name = package['name'].replace('-', '_'))]:
			abspath = os.path.join(basedir, relpath)
			
			if os.path.exists(abspath):
				if os.path.isfile(abspath):
					os.remove(abspath)
				else:
					shutil.rmtree(abspath)

		for dirpath, dirnames, filenames in os.walk(basedir):
			for filename in filenames:
				_, extension = os.path.splitext(filename)
				
				if extension in ['.pyc']:
					abspath = os.path.join(dirpath, filename)
					os.remove(abspath)
				
			for dirname in dirnames:
				if dirname in ['__pycache__']:
					abspath = os.path.join(dirpath,  dirname)
					shutil.rmtree(abspath)

class TestCommand(Command):
	pass

def main(args = None):
	code = os.EX_OK

	meta = dict(
		name             = package['name'],
		version          = package['version'],
		description      = package['description'],
		url              = package['homepage'],
		author           = ', '.join([author['name']  for author in package['authors']]),
		author_email     = ', '.join([author['email'] for author in package['authors']]),
		long_description = package['long_description'],
		license          = package['license'],
		classifiers      = package['classifiers'],
		keywords         = ' '.join([keyword for keyword in package['keywords']]),
		packages         = find_packages(exclude = ['test']),
		install_requires = package['dependencies']['production'],
		entry_points     = {
			'console_scripts': [
				'bench = bench.cli:cli' # should be main
			]
		},
		cmdclass         = \
		{
			'clean': CleanCommand
		}
	)

	setup(**meta)

	return code

if __name__ == '__main__':
	args = sys.argv[1:]
	code = main(args)

	sys.exit(code)


import os, getpass, click
import bench

def generate_supervisor_config(bench_path, user=None, yes=False):
	from bench.app import get_current_frappe_version, use_rq
	from bench.utils import get_bench_name, find_executable
	from bench.config.common_site_config import get_config, update_config, get_gunicorn_workers

	template = bench.env.get_template('supervisor.conf')
	if not user:
		user = getpass.getuser()

	config = get_config(bench_path=bench_path)

	bench_dir = os.path.abspath(bench_path)

	config = template.render(**{
		"bench_dir": bench_dir,
		"sites_dir": os.path.join(bench_dir, 'sites'),
		"user": user,
		"frappe_version": get_current_frappe_version(bench_path),
		"use_rq": use_rq(bench_path),
		"http_timeout": config.get("http_timeout", 120),
		"redis_server": find_executable('redis-server'),
		"node": find_executable('node') or find_executable('nodejs'),
		"redis_cache_config": os.path.join(bench_dir, 'config', 'redis_cache.conf'),
		"redis_socketio_config": os.path.join(bench_dir, 'config', 'redis_socketio.conf'),
		"redis_queue_config": os.path.join(bench_dir, 'config', 'redis_queue.conf'),
		"webserver_port": config.get('webserver_port', 8000),
		"gunicorn_workers": config.get('gunicorn_workers', get_gunicorn_workers()["gunicorn_workers"]),
		"bench_name": get_bench_name(bench_path),
		"background_workers": config.get('background_workers') or 1,
		"bench_cmd": find_executable('bench')
	})

	conf_path = os.path.join(bench_path, 'config', 'supervisor.conf')
	if not yes and os.path.exists(conf_path):
		click.confirm('supervisor.conf already exists and this will overwrite it. Do you want to continue?',
			abort=True)

	with open(conf_path, 'w') as f:
		f.write(config)

	update_config({'restart_supervisor_on_update': True}, bench_path=bench_path)




import os, sys, json, click
from bench.utils import run_playbook, setup_sudoers

extra_vars = {"production": True}

@click.group()
def install():
	"Install system dependancies"
	pass

@click.command('prerequisites')
def install_prerequisites():
	"Install prerequisites"
	run_playbook('prerequisites/install_prerequisites.yml')

@click.command('mariadb')
@click.option('--mysql_root_password')
def install_maridb(mysql_root_password=''):
	"Install mariadb 10.1"
	if mysql_root_password:
		extra_vars.update({"mysql_root_password": mysql_root_password})

	run_playbook('prerequisites/install_roles.yml', extra_vars=extra_vars, tag='mariadb')

@click.command('wkhtmltopdf')
def install_wkhtmltopdf():
	run_playbook('prerequisites/install_roles.yml', extra_vars=extra_vars, tag='wkhtmltopdf')

@click.command('nodejs')
def install_nodejs():
	run_playbook('prerequisites/install_roles.yml', extra_vars=extra_vars, tag='nodejs')

@click.command('psutil')
def install_psutil():
	run_playbook('prerequisites/install_roles.yml', extra_vars=extra_vars, tag='psutil')

@click.command('supervisor')
@click.option('--user')
def install_supervisor(user=None):
	run_playbook('prerequisites/install_roles.yml', extra_vars=extra_vars, tag='supervisor')
	if user:
		setup_sudoers(user)

@click.command('nginx')
@click.option('--user')
def install_nginx(user=None):
	run_playbook('prerequisites/install_roles.yml', extra_vars=extra_vars, tag='nginx')
	if user:
		setup_sudoers(user)

@click.command('fail2ban')
def install_failtoban():
	run_playbook('prerequisites/install_roles.yml', extra_vars=extra_vars, tag='fail2ban')

install.add_command(install_prerequisites)
install.add_command(install_maridb)
install.add_command(install_wkhtmltopdf)
install.add_command(install_nodejs)
install.add_command(install_psutil)
install.add_command(install_supervisor)
install.add_command(install_nginx)
install.add_command(install_failtoban)


import click

def print_bench_version(ctx, param, value):
	"""Prints current bench version"""
	if not value or ctx.resilient_parsing:
		return

	import bench
	click.echo(bench.__version__)
	ctx.exit()

@click.group()
@click.option('--version', is_flag=True, is_eager=True, callback=print_bench_version, expose_value=False)
def bench_command(bench_path='.'):
	"""Bench manager for Frappe"""
	import bench
	from bench.app import get_current_frappe_version
	from bench.utils import setup_logging

	bench.set_frappe_version(bench_path=bench_path)
	setup_logging(bench_path=bench_path)


from bench.commands.make import init, get_app, new_app, remove_app, new_site
bench_command.add_command(init)
bench_command.add_command(get_app)
bench_command.add_command(new_app)
bench_command.add_command(remove_app)
bench_command.add_command(new_site)


from bench.commands.update import update, retry_upgrade, switch_to_branch, switch_to_master, switch_to_develop
bench_command.add_command(update)
bench_command.add_command(retry_upgrade)
bench_command.add_command(switch_to_branch)
bench_command.add_command(switch_to_master)
bench_command.add_command(switch_to_develop)

from bench.commands.utils import (start, restart, set_nginx_port, set_ssl_certificate, set_ssl_certificate_key, set_url_root,
	set_mariadb_host, set_default_site, download_translations, shell, backup_site, backup_all_sites, release, renew_lets_encrypt,
	disable_production, bench_src, prepare_staging)
bench_command.add_command(start)
bench_command.add_command(restart)
bench_command.add_command(set_nginx_port)
bench_command.add_command(set_ssl_certificate)
bench_command.add_command(set_ssl_certificate_key)
bench_command.add_command(set_url_root)
bench_command.add_command(set_mariadb_host)
bench_command.add_command(set_default_site)
bench_command.add_command(download_translations)
bench_command.add_command(shell)
bench_command.add_command(backup_site)
bench_command.add_command(backup_all_sites)
bench_command.add_command(release)
bench_command.add_command(prepare_staging)
bench_command.add_command(renew_lets_encrypt)
bench_command.add_command(disable_production)
bench_command.add_command(bench_src)

from bench.commands.setup import setup
bench_command.add_command(setup)


from bench.commands.config import config
bench_command.add_command(config)

from bench.commands.git import remote_set_url, remote_reset_url, remote_urls
bench_command.add_command(remote_set_url)
bench_command.add_command(remote_reset_url)
bench_command.add_command(remote_urls)

from bench.commands.install import install
bench_command.add_command(install)